"""Constants used for internally processing file chunks and for some default parameter values"""


class DataFileHandlingConstants:
    """
    Constants for internally handling DataFileChunks
    """

    CHUNK_ALREADY_WRITTEN_CODE = 10
    FILE_HASH_MISMATCH_CODE = -1
    FILE_SUCCESSFULLY_RECONSTRUCTED_CODE = 3
    FILE_IN_PROGRESS = 2


DATA_FILE_HANDLING_CONST = DataFileHandlingConstants()



"""Small class to effect transfer of a datafile to an S3 bucket"""

from botocore.exceptions import ClientError
from .s3_service import S3Service


class S3DataTransfer(S3Service):
    """
    Small class to effect transfer of a datafile to an S3 bucket
    """

    def __init__(self, s3_config, *args, **kwargs):
        super().__init__(s3_config, *args, **kwargs)

    def transfer_object_stream(self, object_key, datafile):
        """
        Transfer the contents of a DataFile Consumed from chunks in a topic to an S3 bucket
        """
        content = None
        try:
            content = datafile.bytestring
        except AttributeError:
            with open(datafile.full_filepath, "rb") as fp:
                content = fp.read()
        file_name = str(datafile.filename)
        try:
            self.s3_client.put_object(
                Body=content,
                Bucket=self.bucket_name,
                Key=object_key,
                # GrantRead=self.grant_read
            )
            msg = f"{file_name} successfully transferred into {self.bucket_name}/{object_key}"
            self.logger.debug(msg)
        except ClientError as err:
            errmsg = (
                f"{err.response}: failed to transfer {file_name} into "
                f"{self.bucket_name}/{object_key}"
            )
            self.logger.error(errmsg, exc_info=err)



"""
Transfer contents of DataFiles read from chunks in a topic to an S3 bucket
when complete files become available
"""

# imports
from ..data_file_io.actor.data_file_stream_processor import DataFileStreamProcessor
from .config_file_parser import S3ConfigFileParser
from .s3_data_transfer import S3DataTransfer


class S3TransferStreamProcessor(DataFileStreamProcessor):
    """
    A class to reconstruct data files read as messages from a topic, hold them in memory
    or on disk, and transfer them to an S3 bucket when all of their messages have been
    received

    :param bucket_name: Name of the S3 bucket into which reconstructed files should be transferred
    :type bucket_name: str
    :param config_path: Path to the config file to use in defining the Broker connection
        and Consumers
    :type config_path: :class:`pathlib.Path`
    :param topic_name: Name of the topic to which the Consumers should be subscribed
    :type topic_name: str
    :param filepath_regex: If given, only messages associated with files whose paths match
        this regex will be consumed
    :type filepath_regex: :type filepath_regex: :func:`re.compile` or None, optional
    """

    def __init__(self, bucket_name, config_path, topic_name, **kwargs):
        super().__init__(config_path, topic_name, **kwargs)
        parser = S3ConfigFileParser(config_path, logger=self.logger)
        self.__s3_config = parser.s3_configs
        self.__s3_config["bucket_name"] = bucket_name
        self.bucket_name = bucket_name
        self.s3d = S3DataTransfer(self.__s3_config, logger=self.logger)

    def make_stream(self):
        """
        Runs :func:`~DataFileStreamProcessor.process_files_as_read` to reconstruct files in memory
        and transfer completed files to the S3 bucket.
        Runs until the user inputs a command to shut it down.

        :return: the total number of messages consumed
        :rtype: int
        :return: the total number of messages processed (registered in memory)
        :rtype: int
        :return: the paths of files successfully transferred to the S3 bucket during the run
        :rtype: list
        """
        return self.process_files_as_read()

    def _process_downloaded_data_file(self, datafile, lock):
        """
        Transfer a fully-reconstructed file to the S3 bucket and verify that its contents
        in the bucket match its original hash from disk.
        Logs a warning if the file hashes don't match.

        :param datafile: A :class:`~DownloadDataFile` object that has received
            all of its messages from the topic
        :type datafile: :class:`~DownloadDataFile`
        :param lock: Acquiring this :class:`threading.Lock` object ensures that only one instance
            of :func:`~_process_downloaded_data_file` is running at once
        :type lock: :class:`threading.Lock`

        :return: None if processing was successful, a caught Exception otherwise
        """
        object_key = self.__get_datafile_object_key(datafile)
        try:
            self.s3d.transfer_object_stream(object_key, datafile)
        except Exception as exc:
            self.logger.error(
                f"ERROR: failed to transfer {datafile.filename} to the object store"
            )
            return exc
        if self.s3d.compare_consumer_datafile_with_s3_object_stream(
            self.bucket_name, object_key, datafile
        ):
            self.logger.debug(object_key + " matched with consumer datafile")
            # self.s3d.delete_object_from_bucket(self.bucket_name, object_key)
        else:
            warnmsg = (
                f"WARNING: {object_key} transferred to bucket but the file in the bucket "
                "does not match the file originally read from disk!"
            )
            self.logger.warning(warnmsg)
        return None

    def __get_datafile_object_key(self, datafile):
        sub_dir = datafile.subdir_str
        return (
            f'{self.topic_name}{"/" if sub_dir!="" else ""}{sub_dir}/{datafile.filename}'
        )

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = [*superargs, "bucket_name"]
        kwargs = {**superkwargs, "config": "test_s3_transfer"}
        return args, kwargs

    @classmethod
    def run_from_command_line(cls, args=None):
        """
        Run a :class:`~S3TransferStreamProcessor` directly from the command line

        Calls :func:`~make_stream` on a :class:`~S3TransferStreamProcessor` defined by
        command line (or given) arguments

        :param args: the list of arguments to send to the parser instead of getting them
            from sys.argv
        :type args: list, optional
        """
        # make the argument parser
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        s3_stream_proc = cls(
            args.bucket_name,
            args.config,
            args.topic_name,
            output_dir=args.output_dir,
            mode=args.mode,
            n_threads=args.n_threads,
            filepath_regex=args.download_regex,
            update_secs=args.update_seconds,
            consumer_group_id=args.consumer_group_id,
            streamlevel=args.logger_stream_level,
            filelevel=args.logger_file_level,
        )
        # cls.bucket_name = args.bucket_name
        msg = (
            f"Listening to the {args.topic_name} topic for files to add to the "
            f"{args.bucket_name} bucket...."
        )
        s3_stream_proc.logger.info(msg)
        n_read, n_processed, n_complete, complete_filenames = s3_stream_proc.make_stream()
        s3_stream_proc.close()
        msg = (
            f"{n_read} total messages were consumed, {n_processed} messages were "
            f"successfully processed, and {n_complete} files were transferred "
            f"to the {args.bucket_name} bucket"
        )
        s3_stream_proc.logger.info(msg)
        if len(complete_filenames) > 0:
            msg = (
                f'{n_complete} file{" was" if n_complete==1 else "s were"} successfully '
                f"transferred to the {args.bucket_name} bucket.\nTransferred filepaths "
                f"(up to {cls.N_RECENT_FILES} most recent):\n\t"
            )
            msg += "\n\t".join(complete_filenames)
            s3_stream_proc.logger.debug(msg)

    def _on_check(self):
        msg = (
            f"{self.n_msgs_read} messages read, {self.n_msgs_processed} messages processed, "
            f"{self.n_processed_files} files transferred so far"
        )
        self.logger.info(msg)
        if (
            len(self.files_in_progress_by_path) > 0
            or len(self.recent_processed_filepaths) > 0
        ):
            self.logger.debug(self.progress_msg)


def main(args=None):
    """
    Main method to run from command line
    """
    S3TransferStreamProcessor.run_from_command_line(args)


if __name__ == "__main__":
    main()



"""Custom argument parser and associated functions"""

# imports
import pathlib, math, re, logging
from argparse import ArgumentParser
from .config import RUN_CONST

#################### MISC. FUNCTIONS ####################


def existing_file(argstring):
    """
    convert a string or path argument into a path to a file, checking if it exists
    """
    filepath = pathlib.Path(argstring)
    if filepath.is_file():
        return filepath.resolve()
    raise FileNotFoundError(f"ERROR: file {argstring} does not exist!")


def existing_dir(argstring):
    """
    convert a string or path argument into a directory path, checking if it exists
    """
    dirpath = pathlib.Path(argstring)
    if dirpath.is_dir():
        return dirpath.resolve()
    raise FileNotFoundError(f"ERROR: directory {argstring} does not exist!")


def create_dir(argstring):
    """
    convert a string or path argument into a directory path, creating it if necessary
    """
    if argstring is None:  # Then the argument wasn't given and nothing should be done
        return None
    dirpath = pathlib.Path(argstring)
    if dirpath.is_dir():
        return dirpath.resolve()
    if dirpath.exists():
        raise RuntimeError(
            f"ERROR: directory path {argstring} exists but is not a directory!"
        )
    dirpath.mkdir(parents=True)
    return dirpath.resolve()


def config_path(configarg):
    """
    convert a string or path argument into a config file path
    (raise an exception if the file can't be found)
    """
    if isinstance(configarg, str) and "." not in configarg:
        configarg += RUN_CONST.CONFIG_FILE_EXT
    configpath = pathlib.Path(configarg)
    if configpath.is_file():
        return configpath.resolve()
    if (RUN_CONST.CONFIG_FILE_DIR / configpath).is_file():
        return (RUN_CONST.CONFIG_FILE_DIR / configpath).resolve()
    raise ValueError(
        f"ERROR: config argument {configarg} is not a recognized config file!"
    )


def detect_bucket_name(argstring):
    """
    detects if the bucket name contains invalid characters
    """
    if argstring is None:  # Then the argument wasn't given and nothing should be done
        return None
    illegal_characters = [
        "#",
        "%",
        "&",
        "{",
        "}",
        "\\",
        "/",
        "<",
        ">",
        "*",
        "?",
        " ",
        "$",
        "!",
        "'",
        '"',
        ":",
        "@",
        "+",
        "`",
        "|",
        "=",
    ]
    for illegal_character in illegal_characters:
        if illegal_character in argstring:
            raise RuntimeError(f"ERROR: Illegal characters in bucket_name {argstring}")
    return argstring


def int_power_of_two(argval):
    """
    make sure a given value is a nonzero integer power of two (or can be converted to one)
    """
    if not isinstance(argval, int):
        argval = int(argval)
    if argval <= 0 or math.ceil(math.log2(argval)) != math.floor(math.log2(argval)):
        raise ValueError(
            f"ERROR: invalid argument: {argval} must be a (nonzero) power of two!"
        )
    return argval


def positive_int(argval):
    """
    make sure a given value is a positive integer
    """
    argval = int(argval)
    if (not isinstance(argval, int)) or (argval < 1):
        raise ValueError(f"ERROR: invalid argument: {argval} must be a positive integer!")
    return argval


def logger_string_to_level(argval):
    """
    converts a given string representing a logger level to its corresponding integer
    """
    argval = argval.lower()
    if argval == "notset":
        return logging.NOTSET
    if argval == "debug":
        return logging.DEBUG
    if argval == "info":
        return logging.INFO
    if argval == "warning":
        return logging.WARNING
    if argval == "error":
        return logging.ERROR
    if argval == "critical":
        return logging.CRITICAL
    try:
        if int(argval) >= 0:
            return int(argval)
        raise ValueError(f"ERROR: logger argument {argval} is not valid!")
    except ValueError as exc:
        raise exc


#################### MYARGUMENTPARSER CLASS ####################


class OpenMSIStreamArgumentParser(ArgumentParser):
    """
    Class to make it easier to get an ArgumentParser with some commonly-used arguments in it.

    All constructor arguments get passed to the underlying :class:`argparse.ArgumentParser` object.

    Arguments for the parser are defined in the :attr:`~OpenMSIStreamArgumentParser.ARGUMENTS`
    class variable, which is a dictionary. The keys are names of arguments, and the values
    are lists. The first entry in each list is a string reading "positional" or "optional"
    depending on the type of argument, and the second entry is a dictionary of keyword arguments
    to send to :func:`argparse.ArgumentParser.add_argument`.
    """

    ARGUMENTS = {
        "filepath": [
            "positional",
            {"type": existing_file, "help": "Path to the data file to use"},
        ],
        "output_dir": [
            "positional",
            {"type": create_dir, "help": "Path to the directory to put output in"},
        ],
        "upload_dir": [
            "positional",
            {
                "type": existing_dir,
                "help": "Path to the directory to watch for files to upload",
            },
        ],
        "bucket_name": [
            "positional",
            {
                "type": detect_bucket_name,
                "help": (
                    "Name of the object store bucket to which consumed files should be transferred."
                    "Name should only contain valid characters"
                ),
            },
        ],
        "config": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_CONFIG_FILE,
                "type": config_path,
                "help": (
                    f"Name of config file to use in {RUN_CONST.CONFIG_FILE_DIR.resolve()}, "
                    "or path to a file in a different location"
                ),
            },
        ],
        "topic_name": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_TOPIC_NAME,
                "help": "Name of the topic to produce to or consume from",
            },
        ],
        "consumer_topic_name": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_TOPIC_NAME,
                "help": "Name of the topic to consume from",
            },
        ],
        "producer_topic_name": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_TOPIC_NAME,
                "help": "Name of the topic to produce to",
            },
        ],
        "n_threads": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_N_THREADS,
                "type": positive_int,
                "help": "Maximum number of threads to use",
            },
        ],
        "n_consumer_threads": [
            "optional",
            {
                "default": RUN_CONST.N_DEFAULT_DOWNLOAD_THREADS,
                "type": positive_int,
                "help": "Number of consumer threads to use",
            },
        ],
        "n_producer_threads": [
            "optional",
            {
                "default": RUN_CONST.N_DEFAULT_UPLOAD_THREADS,
                "type": positive_int,
                "help": "Number of producer threads to use",
            },
        ],
        "upload_regex": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_UPLOAD_REGEX,
                "type": re.compile,
                "help": "Only files with paths matching this regular expression will be uploaded",
            },
        ],
        "download_regex": [
            "optional",
            {
                "type": re.compile,
                "help": "Only files with paths matching this regular expression will be uploaded",
            },
        ],
        "chunk_size": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_CHUNK_SIZE,
                "type": int_power_of_two,
                "help": (
                    "Max size (in bytes) of chunks into which files should be broken "
                    "as they are uploaded"
                ),
            },
        ],
        "queue_max_size": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_MAX_UPLOAD_QUEUE_MEGABYTES,
                "type": positive_int,
                "help": (
                    "Maximum allowed size in MB of the internal upload queue. "
                    "Use to adjust RAM usage if necessary."
                ),
            },
        ],
        "update_seconds": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_UPDATE_SECONDS,
                "type": int,
                "help": (
                    'Number of seconds between printing a "." to the console '
                    "to indicate the program is alive"
                ),
            },
        ],
        "mode": [
            "optional",
            {
                "choices": ["memory", "disk", "both"],
                "default": "memory",
                "help": (
                    "Choose whether files should be reconstructed in 'memory', "
                    "on 'disk' (with the output directory as the root directory) "
                    "or 'both' (giving access to the data bytestring, "
                    "but also writing to disk)"
                ),
            },
        ],
        "upload_existing": [
            "optional",
            {
                "action": "store_true",
                "help": (
                    "Add this flag to upload files already existing in addition to those "
                    "added to the directory after this code starts running (by default "
                    "only files added after startup will be uploaded)"
                ),
            },
        ],
        "watchdog_lag_time": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_WATCHDOG_LAG_TIME,
                "type": int,
                "help": (
                    "Number of seconds that a file must remain static (unmodified) "
                    "for its upload to begin"
                ),
            },
        ],
        "consumer_group_id": [
            "optional",
            {"default": "create_new", "help": "ID to use for all consumers in the group"},
        ],
        "optional_output_dir": [
            "optional",
            {"type": create_dir, "help": "Optional path to directory to put output in"},
        ],
        "service_name": ["positional", {"help": "The name of the service to work with"}],
        "optional_service_name": [
            "optional",
            {
                "help": (
                    "The customized name of the Service that will be installed "
                    "to run the chosen class"
                )
            },
        ],
        "run_mode": [
            "positional",
            {
                "choices": [
                    "start",
                    "status",
                    "stop",
                    "remove",
                    "stop_and_remove",
                    "reinstall",
                    "stop_and_reinstall",
                    "stop_and_restart",
                ],
                "help": "What to do with the service",
            },
        ],
        "remove_env_vars": [
            "optional",
            {
                "action": "store_true",
                "help": (
                    "Add this flag to also remove username/password environment variables"
                    "when removing a Service"
                ),
            },
        ],
        "remove_install_args": [
            "optional",
            {
                "action": "store_true",
                "help": (
                    "Add this flag to also remove the install arguments file "
                    "when removing a Service"
                ),
            },
        ],
        "remove_nssm": [
            "optional",
            {
                "action": "store_true",
                "help": "Add this flag to also remove the NSSM executable when removing a Service",
            },
        ],
        "logger_stream_level": [
            "optional",
            {
                "type": logger_string_to_level,
                "default": "info",
                "help": (
                    "Messages below this level will not be processed "
                    "by the logger's stream handler"
                ),
            },
        ],
        "logger_file_level": [
            "optional",
            {
                "type": logger_string_to_level,
                "default": "warning",
                "help": (
                    "Messages below this level will not be processed by the logger's file handler"
                ),
            },
        ],
        "girder_api_url": [
            "positional",
            {
                "help": (
                    "The full path to the REST API of a Girder instance, "
                    "e.g. http://my.girder.com/api/v1."
                )
            },
        ],
        "girder_api_key": [
            "positional",
            {"help": "The API key to use for authenticating to the Girder instance"},
        ],
        "girder_root_folder_id": [
            "optional",
            {
                "help": (
                    "The ID of the Girder Folder relative to which files should be "
                    "uploaded. If this argument is given, it will supersede both of the "
                    "'collection_name' and 'girder_root_folder_path' arguments."
                ),
            },
        ],
        "collection_name": [
            "optional",
            {
                "help": (
                    "The name of the top-level Collection to which files should be uploaded. "
                    "Superseded if 'girder_root_folder_id' is given."
                ),
            },
        ],
        "girder_root_folder_path": [
            "optional",
            {
                "help": (
                    "The name of the Folder inside the top-level Collection relative to "
                    "which files should be uploaded. A path to a subdirectory in the "
                    "Collection can be given using forward slashes. "
                    "Superseded if 'girder_root_folder_id' is given. "
                    "(default = collection_name/topic_name)"
                ),
            },
        ],
        "provider": [
            "optional",
            {
                "help": (
                    'Adding this argument will add a corresponding "provider" metadata '
                    "field to all created folders and items"
                ),
            },
        ],
    }

    #################### OVERLOADED FUNCTIONS ####################

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__argnames_added = []
        self.__subparsers_action_obj = None
        self.__subparsers = {}
        self.__subparser_argnames_added = {}

    def add_subparsers(self, *args, **kwargs):
        """
        Overloaded from base class; OpenMSIStreamArgumentParser actually owns its subparsers
        to simplify conflicting argument names
        """
        if self.__subparsers_action_obj is not None or self.__subparsers:
            raise RuntimeError(
                "ERROR: add_subparsers called for an argument parser that already has subparsers!"
            )
        self.__subparsers_action_obj = super().add_subparsers(*args, **kwargs)

    def parse_args(self, *args, **kwargs):
        """
        Overloaded from base class to print usage when catching errors in parsing arguments
        """
        try:
            return super().parse_args(*args, **kwargs)
        except Exception:
            self.print_usage()
            raise

    #################### UNIQUE FUNCTIONS ####################

    def add_arguments(self, *args, **kwargs):
        """
        Add a group of common arguments to the parser

        :param args: Names of arguments that should be added exactly as they're listed in
            :attr:`~OpenMSIStreamArgumentParser.ARGUMENTS`
        :type args: list
        :param kwargs: Dictionary whose keys are argument names like in `args` and
            whose values are new default argument values
        :type kwargs: dict

        :raises ValueError: if the name of an argument isn't recognized in
            :attr:`~OpenMSIStreamArgumentParser.ARGUMENTS`, or if the type of a new
            default argument is different than the type of its original default argument
        """
        if len(args) < 1 and len(kwargs) < 1:
            raise ValueError(
                "ERROR: must specify at least one desired argument to create an argument parser!"
            )
        args_to_use = []
        for argname in args:
            if argname not in args_to_use:
                args_to_use.append(argname)
        for argname in kwargs:
            if argname in args_to_use:
                args_to_use.remove(argname)
        for argname in args_to_use:
            argname_to_add, kwargs_for_arg = self.__get_argname_and_kwargs(argname)
            if argname_to_add not in self.__argnames_added:
                self.add_argument(argname_to_add, **kwargs_for_arg)
                self.__argnames_added.append(argname_to_add)
        for argname, argdefault in kwargs.items():
            argname_to_add, kwargs_for_arg = self.__get_argname_and_kwargs(
                argname, argdefault
            )
            if argname_to_add not in self.__argnames_added:
                self.add_argument(argname_to_add, **kwargs_for_arg)
                self.__argnames_added.append(argname_to_add)

    def add_subparser_arguments(
        self, subp_name, args_to_add=None, kwargs_to_add=None, **other_kwargs
    ):
        """
        Create a new subparser and add arguments to it.

        :param subp_name: the name of the subparser command
        :type subp_name: str
        :param args_to_add: additional arguments that should be added to the subparser
            (same format as `args` for :func:`~OpenMSIStreamArgumentParser.add_arguments`)
        :type args_to_add: list, optional
        :param kwargs_to_add: additional keyword arguments that should be added to the subparser
            (same format as `kwargs` for :func:`~OpenMSIStreamArgumentParser.add_arguments`)
        :type kwargs_to_add: dict, optional
        :param other_kwargs: any other keyword arguments are passed to the subparser's
            add_parser() method
        :type other_kwargs: dict, optional

        :raises ValueError: like in :func:`~OpenMSIStreamArgumentParser.add_arguments`
        """
        if self.__subparsers_action_obj is None:
            errmsg = (
                "ERROR: add_subparser_arguments called for an argument parser "
                "that has not added subparsers!"
            )
            raise RuntimeError(errmsg)
        if subp_name in self.__subparsers:
            errmsg = (
                f"ERROR: subparser arguments for {subp_name} have already been added "
                "to this argument parser!"
            )
            raise RuntimeError(errmsg)
        self.__subparsers[subp_name] = self.__subparsers_action_obj.add_parser(
            subp_name, **other_kwargs
        )
        self.__subparser_argnames_added[subp_name] = []
        if args_to_add is not None:
            for argname in args_to_add:
                argname_to_add, kwargs_for_arg = self.__get_argname_and_kwargs(argname)
                if argname_to_add not in self.__subparser_argnames_added[subp_name]:
                    self.__subparsers[subp_name].add_argument(
                        argname_to_add, **kwargs_for_arg
                    )
                    self.__subparser_argnames_added[subp_name].append(argname_to_add)
        if kwargs_to_add is not None:
            for argname, argdefault in kwargs_to_add.items():
                argname_to_add, kwargs_for_arg = self.__get_argname_and_kwargs(
                    argname, argdefault
                )
                if argname_to_add not in self.__subparser_argnames_added[subp_name]:
                    self.__subparsers[subp_name].add_argument(
                        argname_to_add, **kwargs_for_arg
                    )
                    self.__subparser_argnames_added[subp_name].append(argname_to_add)

    def add_subparser_arguments_from_class(
        self,
        class_to_add,
        *,
        subp_name=None,
        addl_args=None,
        addl_kwargs=None,
        **other_kwargs,
    ):
        """
        Create a new subparser and add arguments from the given class to it.
        `class_to_add must` inherit from :class:`~Runnable` to be able to get its arguments.

        :param subp_name: an override for the name of the command the subparser
            should be registered under. (Default is the name of the class.)
        :type subp_name: str, optional
        :param addl_args: additional arguments that should be added to the subparser
            (same format as `args` for :func:`~OpenMSIStreamArgumentParser.add_arguments`)
        :type addl_args: list, optional
        :param addl_kwargs: additional keyword arguments that should be added to the subparser
            (same format as `kwargs` for :func:`~OpenMSIStreamArgumentParser.add_arguments`)
        :type addl_kwargs: dict, optional
        :param other_kwargs: any other keyword arguments are passed to subparsers'
            add_parser() method
        :type other_kwargs: dict, optional

        :raises ValueError: like in :func:`~OpenMSIStreamArgumentParser.add_arguments`
        """
        if subp_name is None:
            subp_name = class_to_add.__name__
        argnames, argnames_with_defaults = class_to_add.get_command_line_arguments()
        if addl_args is not None:
            argnames = [*argnames, *addl_args]
        if addl_kwargs is not None:
            argnames_with_defaults = {**argnames_with_defaults, **addl_kwargs}
        self.add_subparser_arguments(
            subp_name, argnames, argnames_with_defaults, **other_kwargs
        )

    def __get_argname_and_kwargs(self, argname, new_default=None):
        """
        Return the name and kwargs dict for a particular argument

        argname = the given name of the argument, will be matched to a name in ARGUMENTS
        new_default = override the default already specified in ARGUMENTS
        """
        if argname in self.ARGUMENTS:
            if self.ARGUMENTS[argname][0] == "positional":
                argname_to_add = argname
            else:
                if argname.startswith("optional_"):
                    argname_to_add = f'--{argname[len("optional_"):]}'
                else:
                    argname_to_add = f"--{argname}"
            kwargs = self.ARGUMENTS[argname][1].copy()
            if new_default is not None:
                if "default" in kwargs.keys() and not isinstance(
                    new_default, type(kwargs["default"])
                ):
                    errmsg = (
                        f"ERROR: new default value {new_default} for argument {argname} "
                        f"is of a different type than expected based on the old default "
                        f'({self.ARGUMENTS[argname]["kwargs"]["default"]})!'
                    )
                    raise ValueError(errmsg)
                kwargs["default"] = new_default
            if "default" in kwargs.keys():
                if "help" in kwargs.keys():
                    kwargs["help"] += f" (default = {kwargs['default']})"
                else:
                    kwargs["help"] = f"default = {kwargs['default']}"
            return argname_to_add, kwargs
        raise ValueError(f"ERROR: argument {argname} is not recognized as an option!")

    @property
    def actions(self):
        """
        wrapper around parser._actions to make it publicly available
        """
        return self._actions



"""Classes for OpenMSIStream logging infrastructure"""

# imports
import pathlib, logging
from .has_arguments import HasArguments


class OpenMSIStreamFormatter(logging.Formatter):
    """
    Very small extension of the usual logging.Formatter to allow modification of format
    based on message content
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def format(self, record):
        """
        If a message starts with a newline, start the actual logging line with the newline
        before any of the rest
        """
        formatted = ""
        if record.msg.startswith("\n"):
            record.msg = record.msg.lstrip("\n")
            formatted += "\n"
        formatted += super().format(record)
        return formatted


class Logger:
    """
    Class for a general logger in OpenMSIStream format.

    :param logger_name: The name for the logger to use
        (automatically inferred from the running module if not given)
    :type logger_name: str, optional
    :param streamlevel: The level at/above which messages should be logged to the stream/console
    :type streamlevel: logging level int, optional
    :param logger_filepath: The path to a logger file to use or directory in which
        an automatically-named logger file should be created
    :type logger_filepath: :class:`pathlib.Path`, optional
    :param filelevel: The level at/above which messages should be written to the logfile
    :type filelevel: logging level int, optional
    """

    FORMATTER = OpenMSIStreamFormatter(
        "[%(name)s %(asctime)s] %(message)s", "%Y-%m-%d %H:%M:%S"
    )

    def __init__(
        self,
        logger_name=None,
        streamlevel=logging.INFO,
        logger_filepath=None,
        filelevel=logging.WARNING,
    ):
        """
        name = the name for this logger to use (probably something like the top module that owns it)
        """
        self._name = logger_name
        if self._name is None:
            self._name = self.__class__.__name__
        self._logger_obj = logging.getLogger(self._name)
        self._logger_obj.setLevel(logging.DEBUG)
        self._streamhandler = logging.StreamHandler()
        self._streamhandler.setLevel(streamlevel)
        self._streamhandler.setFormatter(self.FORMATTER)
        self._logger_obj.addHandler(self._streamhandler)
        self._filehandler = None
        if logger_filepath is not None:
            self.add_file_handler(logger_filepath, level=filelevel)

    def set_level(self, level):
        """
        Set the level of the entire underlying logger

        :param level: The level to set
        :type level: logging level int
        """
        self._logger_obj.setLevel(level)

    def set_stream_level(self, level):
        """
        Set the level of the underlying logger's streamhandler

        :param level: The level to set
        :type level: logging level int
        """
        self._streamhandler.setLevel(level)

    def get_stream_level(self):
        """
        Get the current level of the underlying logger's streamhandler

        :return: The integer level of the current streamhandler
        """
        return self._streamhandler.level

    def set_file_level(self, level):
        """
        Set the level of the underlying logger's filehandler

        :param level: The level to set
        :type level: logging level int
        """
        if self._filehandler is None:
            errmsg = (
                f"ERROR: Logger {self._name} does not have a filehandler set "
                "but set_file_level was called!"
            )
            raise RuntimeError(errmsg)
        self._filehandler.setLevel(level)

    def add_file_handler(self, filepath, level=logging.INFO):
        """
        Add an additional :class:`logging.FileHandler` to the logger

        :param filepath: The path to the new logger file
        :type filepath: :class:`pathlib.Path`
        :param level: The level to set
        :type level: logging level int
        """
        if not isinstance(filepath, pathlib.PurePath):
            self.error(
                f"ERROR: {filepath} is a {type(filepath)} object, not a Path object!",
                exc_type=TypeError,
            )
        if not filepath.is_file():
            if not filepath.parent.is_dir():
                filepath.parent.mkdir(parents=True)
            filepath.touch()
        self._filehandler = logging.FileHandler(filepath)
        self._filehandler.setLevel(level)
        self._filehandler.setFormatter(self.FORMATTER)
        self._logger_obj.addHandler(self._filehandler)

    # methods for logging different levels of messages

    def debug(self, msg, *args, **kwargs):
        """
        Log a message at DEBUG level. Additional args/kwargs are sent to
        the underlying logger object's debug call.

        :param msg: the message to log
        :type msg: str
        """
        self._logger_obj.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        Log a message at INFO level. Additional args/kwargs are sent to
        the underlying logger object's info call.

        :param msg: the message to log
        :type msg: str
        """
        self._logger_obj.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        Log a message at WARNING level. Additional args/kwargs are sent to
        the underlying logger object's warning call.

        :param msg: the message to log (will have "WARNING: " prepended if it doesn't start with it)
        :type msg: str
        """
        if not msg.startswith("WARNING:"):
            msg = f"WARNING: {msg}"
        self._logger_obj.warning(msg, *args, **kwargs)

    def error(self, msg, *, exc_type=None, reraise=False, **kwargs):
        """
        Log a message at ERROR level. Optionally raise an exception of a given type
        with the same message, or re-raise an Exception object passed through the
        `exc_info` kwarg (after logging its traceback).

        Additional kwargs are sent to the underlying logger object's error call.

        :param msg: the message to log
        :type msg: str
        :param exc_type: The type of Exception to raise with the same message as `msg`
        :type exc_type: :class:`Exception`, optional
        :param reraise: if True, any :class:`Exception` object passed through the `exc_info`
            will be re-raised after its traceback is logged.
        :type reraise: bool, optional
        """
        if not msg.startswith("ERROR:"):
            msg = f"ERROR: {msg}"
        self._logger_obj.error(msg, **kwargs)
        if (
            reraise
            and ("exc_info" in kwargs)
            and isinstance(kwargs["exc_info"], Exception)
        ):
            raise kwargs["exc_info"]
        if exc_type is not None:
            raise exc_type(msg)


class LogOwner(HasArguments):
    """
    Any subclasses extending this one will have access to a Logger defined by
    the first class in the MRO to extend it

    :param logger: a :class:`~.utilities.Logger` object that this class should have access to.
        If this parameter is given it will override any of the others provided.
    :type logger: :class:`~.utilities.Logger`, optional
    :param logger_name: The name for the logger to use
        (automatically inferred from the running module if not given)
    :type logger_name: str, optional
    :param streamlevel: The level at/above which messages should be logged to the stream/console
    :type streamlevel: logging level int, optional
    :param logger_filepath: The path to a logger file to use or directory in which
        an automatically-named logger file should be created
    :type logger_filepath: :class:`pathlib.Path`, optional
    :param filelevel: The level at/above which messages should be written to the logfile
    :type filelevel: logging level int, optional
    """

    @property
    def logger(self):
        """
        The logger object that the class can use
        """
        return self.__logger

    @logger.setter
    def logger(self, logger):
        if (
            hasattr(self, "_LogOwner__logger")
            and self.__logger is not None
            and (not isinstance(self.__logger, type(logger)))
        ):
            errmsg = (
                f"ERROR: tried to reset a logger of type {type(self.__logger)} "
                f"to a new logger of type {logger}!"
            )
            self.__logger.error(errmsg, exc_type=ValueError)
        else:
            self.__logger = logger

    def __init__(
        self,
        *args,
        logger=None,
        logger_name=None,
        streamlevel=logging.INFO,
        logger_file=None,
        filelevel=logging.WARNING,
        **other_kwargs,
    ):
        if logger is not None:
            self.__logger = logger
        else:
            if logger_name is None:
                logger_name = self.__class__.__name__
            logger_filepath = logger_file
            if logger_file is not None and logger_file.is_dir():
                logger_filepath = logger_file / f"{self.__class__.__name__}.log"
            self.__logger = Logger(logger_name, streamlevel, logger_filepath, filelevel)
        super().__init__(*args, **other_kwargs)

    @classmethod
    def get_command_line_arguments(cls):
        """
        Return the names of arguments for the logger stream and file levels.
        """
        superargs, superkwargs = super().get_command_line_arguments()
        return [*superargs, "logger_stream_level", "logger_file_level"], superkwargs



"""
A directory holding files that may or may not be marked for uploading.
Updates when new files are added.
"""

# imports
import datetime, time
from threading import Lock
from queue import Queue
from watchdog.observers import Observer
from ...kafka_wrapper import ProducerGroup
from ...utilities.config import RUN_CONST
from ...utilities import Runnable, ControlledProcessSingleThread
from ...utilities.misc import populated_kwargs
from ...utilities.exception_tracking_thread import ExceptionTrackingThread
from .. import DataFileDirectory
from .file_registry.producer_file_registry import ProducerFileRegistry
from ..entity.upload_data_file import UploadDataFile
from ..entity.upload_directory_event_handler import UploadDirectoryEventHandler


class DataFileUploadDirectory(
    DataFileDirectory, ControlledProcessSingleThread, ProducerGroup, Runnable
):
    """
    Class representing a directory being watched for new files to be added.
    Files added to the directory are broken into chunks and uploaded as messages to a Kafka topic.

    :param dirpath: Path to the directory that should be watched for new files
    :type dirpath: :class:`pathlib.Path`
    :param config_path: Path to the config file to use in defining the Broker connection
        and Producers
    :type config_path: :class:`pathlib.Path`
    :param upload_regex: only files matching this regular expression will be uploaded
    :type upload_regex: :func:`re.compile`, optional
    :param datafile_type: the type of data file that recognized files should be uploaded as
        (must be a subclass of :class:`~UploadDataFile`)
    :type datafile_type: :class:`~UploadDataFile`, optional
    :param watchdog_lag_time: Number of seconds that files must remain static (unmodified)
        for their uploads to begin
    :type watchdog_lag_time: int

    :raises ValueError: if `datafile_type` is not a subclass of :class:`~UploadDataFile`
    """

    #################### CONSTANTS ####################

    # starting point for how long to wait between pinging the directory looking for new files
    MIN_WAIT_TIME = 0.005
    # never wait more than a minute to check again for new files
    MAX_WAIT_TIME = 60
    # amount of time to wait for watchdog to pull events from the queue
    WATCHDOG_OBSERVER_TIMEOUT = 0.1
    # the max number of files to report status messages for
    N_STATUS_MESSAGE_FILES = 50
    # name of the logging subdirectory
    LOG_SUBDIR_NAME = "LOGS"

    #################### PUBLIC FUNCTIONS ####################

    def __init__(
        self,
        dirpath,
        config_path,
        upload_regex=RUN_CONST.DEFAULT_UPLOAD_REGEX,
        datafile_type=UploadDataFile,
        watchdog_lag_time=RUN_CONST.DEFAULT_WATCHDOG_LAG_TIME,
        **kwargs,
    ):
        """
        Constructor method
        """
        # create a subdirectory for the logs
        self._logs_subdir = dirpath / self.LOG_SUBDIR_NAME
        if not self._logs_subdir.is_dir():
            self._logs_subdir.mkdir(parents=True)
        # put the log file in the subdirectory
        kwargs = populated_kwargs(kwargs, {"logger_file": self._logs_subdir})
        super().__init__(dirpath, config_path, **kwargs)
        if not issubclass(datafile_type, UploadDataFile):
            errmsg = (
                "ERROR: DataFileUploadDirectory requires a datafile_type that is a "
                f"subclass of UploadDataFile but {datafile_type} was given!"
            )
            self.logger.error(errmsg, exc_type=ValueError)
        self.__datafile_type = datafile_type
        self.__wait_time = self.MIN_WAIT_TIME
        self.__lock = Lock()
        self.__observer = Observer(timeout=self.WATCHDOG_OBSERVER_TIMEOUT)
        self.__event_handler = UploadDirectoryEventHandler(
            upload_regex=upload_regex,
            logs_subdir=self._logs_subdir,
            lag_time=watchdog_lag_time,
            logger=self.logger,
        )
        self.__active_files_by_path = {}
        self.__status_message_files = {}
        self.__topic_name = None
        self.__chunk_size = None
        self.__file_registry = None
        self.__upload_queue = None
        self.__producers = []
        self.__upload_threads = []

    def upload_files_as_added(
        self,
        topic_name,
        n_threads=RUN_CONST.N_DEFAULT_UPLOAD_THREADS,
        chunk_size=RUN_CONST.DEFAULT_CHUNK_SIZE,
        max_queue_size=RUN_CONST.DEFAULT_MAX_UPLOAD_QUEUE_MEGABYTES,
        upload_existing=False,
    ):
        """
        Watch for new files to be added to the directory.
        Chunk and produce newly added files as messages to a Kafka topic.

        :param topic_name: Name of the topic to produce messages to
        :type topic_name: str
        :param n_threads: The number of threads to use to produce from the shared queue
        :type n_threads: int, optional
        :param chunk_size: The size of each file chunk in bytes
        :type chunk_size: int, optional
        :param max_queue_size: The maximum size in MB of the internal upload queue
            (this is distinct from the librdkafka buffering queue)
        :type max_queue_size: int, optional
        :param upload_existing: True if any files that already exist in the directory
            should be uploaded. If False (the default) then only files added to the
            directory after startup will be enqueued to the producer
        :type upload_existing: bool, optional
        """
        # set the topic name and chunk size
        self.__topic_name = topic_name
        self.__chunk_size = chunk_size
        msg = (
            f'Will upload {"files in" if upload_existing else "new files added to"}'
            f"{self.dirpath} to the {self.__topic_name} topic as "
            f"{self.__chunk_size}-byte chunks using {n_threads} threads"
        )
        self.logger.info(msg)
        # start up a file registry in the watched directory
        self.__file_registry = ProducerFileRegistry(
            dirpath=self._logs_subdir, topic_name=topic_name, logger=self.logger
        )
        # start the upload queue
        n_max_queue_items = int((1000000 * max_queue_size) / self.__chunk_size)
        self.__upload_queue = Queue(n_max_queue_items)
        # start the producers and upload threads
        for _ in range(n_threads):
            self.__producers.append(self.get_new_producer())
            thread = ExceptionTrackingThread(
                target=self.__producers[-1].produce_from_queue_looped,
                args=(self.__upload_queue, self.__topic_name),
                kwargs={"callback": self.producer_callback},
            )
            thread.start()
            self.__upload_threads.append(thread)
        # if we're going to upload files that are already in the directory, scrape it now
        if upload_existing:
            self.__scrape_dir_for_files()
        # add or modify datafiles to upload based on the contents of the file registry
        self.__startup_from_file_registry()
        # schedule and start the observer
        self.__observer.schedule(self.__event_handler, self.dirpath, recursive=True)
        self.__observer.start()
        # loop until the user inputs a command to stop
        self.run()
        # return a list of relative filepaths that have been uploaded
        to_return = []
        for rel_filepath in self.__file_registry.get_completed_filepaths():
            to_return.append(rel_filepath)
        return to_return

    def producer_callback(
        self, err, msg, prodid, filename, filepath, n_total_chunks, chunk_i
    ):
        """
        A reference to this method is given as the callback for each call to
        :func:`confluent_kafka.Producer.produce`. It is called for every message
        upon acknowledgement by the broker, and it uses the file registries in the LOGS
        subdirectory to keep the information about what has and hasn't been uploaded
        current with what has been received by the broker.

        If the message is the final one to be acknowledged from its corresponding
        :class:`~UploadDataFile`, the :class:`~UploadDataFile` is registered as "fully produced".

        :param err: The error object for the message
        :type err: :class:`confluent_kafka.KafkaError`
        :param msg: The message object
        :type msg: :class:`confluent_kafka.Message`
        :param prodid: The ID of the producer that produced the message
            (hex(id(producer)) in memory)
        :type prodid: str
        :param filename: The name of the file the message is coming from
        :type filename: str
        :param filepath: The full path to the file the message is coming from
        :type filepath: :class:`pathlib.Path`
        :param n_total_chunks: The total number of chunks in the file this message came from
        :type n_total_chunks: int
        :param chunk_i: The index of the message's file chunk in the full list for the file
        :type chunk_i: int
        """
        # If any error occured, re-enqueue the message's file chunk
        if err is None and msg.error() is not None:
            err = msg.error()
        if err is not None:
            warnmsg = (
                f'WARNING: Producer with ID {prodid} {"fatally" if err.fatal() else ""} '
                f"failed to deliver message for chunk {chunk_i} of {filepath}"
                f'{" and cannot retry" if not err.retriable() else ""}. '
                f"This message will be re-enqeued. Error reason: {err.str()}"
            )
            self.logger.warning(warnmsg)
            with self.__lock:
                self.__add_chunks_for_filepath(filepath, [chunk_i])
        # Otherwise, register the chunk as successfully sent to the broker
        else:
            rel_filepath = filepath.relative_to(self.dirpath)
            fully_produced = self.__file_registry.register_chunk(
                filename, rel_filepath, n_total_chunks, chunk_i, prodid
            )
            # If the file has now been fully produced to the topic,
            # set the variable for the file and log a line
            if fully_produced:
                with self.__lock:
                    self.__active_files_by_path[filepath].fully_enqueued = False
                    self.__active_files_by_path[filepath].fully_produced = True
                    self.__forget_inactive_files()
                debugmsg = (
                    f'{rel_filepath} has been fully produced to the "{self.__topic_name}"'
                    f' topic as {n_total_chunks} message{"s" if n_total_chunks!=1 else ""}'
                )
                self.logger.debug(debugmsg)

    def filepath_should_be_uploaded(self, filepath):
        """
        Returns True if a given filepath should be uploaded (if it matches the upload regex)

        :param filepath: A candidate upload filepath
        :type filepath: :class:`pathlib.Path`

        :return: True if the filepath is an existing file outside of the LOGS directory
            whose path matches the upload regex, False otherwise
        :rtype: bool

        :raises TypeError: if `filepath` isn't a :class:`pathlib.Path` object
        """
        return self.__event_handler.filepath_matched(filepath)

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _run_iteration(self):
        """
        If there are any files currently uploading, finish uploading them.
        If not, consult the watchdog event handler to get new active files.
        On iterations where nothing is being actively uploaded, poll the producers
        and increase the timeout for getting event handler updates exponentially
        if nothing is happening.
        """
        # pull any new files from the event handler and poll the producers
        if not self.have_file_to_upload:
            # wait here so that this loop isn't constantly consuming CPU if nothing
            # is happening
            time.sleep(self.__wait_time)
            self.__pull_from_handler()
            n_new_callbacks = 0
            for producer in self.__producers:
                new_callbacks = producer.poll(0)
                if new_callbacks is not None:
                    n_new_callbacks += new_callbacks
            if (not self.have_file_to_upload) and (n_new_callbacks == 0):
                if self.__wait_time < self.MAX_WAIT_TIME:
                    self.__wait_time *= 1.05
            else:
                self.__wait_time = self.MIN_WAIT_TIME
            return
        # find the first file that's running and add some of its chunks to the upload queue
        for datafile in self.__active_files_by_path.values():
            if datafile.upload_in_progress or datafile.waiting_to_upload:
                datafile.enqueue_chunks_for_upload(
                    self.__upload_queue,
                    n_threads=len(self.__upload_threads),
                    chunk_size=self.__chunk_size,
                )
                break
        # restart any crashed threads
        self.__restart_crashed_threads()

    def _on_check(self):
        # poll the producers
        for producer in self.__producers:
            producer.poll(0)
        # forget inactive files
        with self.__lock:
            self.__forget_inactive_files()
        # log progress so far
        self.logger.info(self.status_msg)
        self.logger.debug(self.progress_msg)
        # reset the wait time
        self.__wait_time = self.MIN_WAIT_TIME

    def _on_shutdown(self):
        self.logger.info(
            "Will quit after all currently enqueued files are done being transferred."
        )
        self.logger.debug(self.progress_msg)
        # shut down the watchdog observer
        self.__observer.stop()
        self.__observer.join()
        # add the remainder of any files currently in progress
        if self.n_partially_done_files > 0:
            msg = "Will finish queueing the following files:\n\t" "\n\t".join(
                [str(_) for _ in self.partially_done_file_paths]
            )
            self.logger.info(msg)
        while self.n_partially_done_files > 0:
            for datafile in self.__active_files_by_path.values():
                if datafile.upload_in_progress:
                    datafile.enqueue_chunks_for_upload(
                        self.__upload_queue,
                        n_threads=len(self.__upload_threads),
                        chunk_size=self.__chunk_size,
                    )
                    break
        # stop the uploading threads by adding "None"s to their queues and joining them
        for thread in self.__upload_threads:
            self.__upload_queue.put(None)
        for thread in self.__upload_threads:
            thread.join()
        self.logger.info(
            "Waiting for all enqueued messages to be delivered (this may take a moment)"
        )
        for producer in self.__producers:
            # don't move on until all enqueued messages have been sent/received
            producer.flush(timeout=-1)
            producer.close()
        self.close()
        self.__file_registry.consolidate_completed_files()

    def __add_active_datafile_for_path(self, filepath):
        """
        Create a new datafile with the given path and add it to the dictionaries of
        active datafiles and datafiles to report for status messages

        filepath = the path to the file that should be added
        """
        new_datafile = self.__datafile_type(
            filepath,
            to_upload=True,
            rootdir=self.dirpath,
            logger=self.logger,
            **self.other_datafile_kwargs,
        )
        self.__active_files_by_path[filepath] = new_datafile
        n_keys_to_remove = (
            len(self.__status_message_files) + 1 - self.N_STATUS_MESSAGE_FILES
        )
        if n_keys_to_remove > 0:
            keys_to_remove = list(self.__status_message_files.keys())[:n_keys_to_remove]
            for key in keys_to_remove:
                _ = self.__status_message_files.pop(key)
        self.__status_message_files[filepath] = new_datafile

    def __scrape_dir_for_files(self, retries_left=10):
        """
        Search the directory for any unrecognized files and add them as active datafiles

        retries_left = how many attempts should be made to retry if an error is encountered
        """
        # This is in a try/except in case a file is moved or a subdirectory is renamed
        # while this method is running: it'll retry up to ten times
        try:
            for filepath in self.dirpath.rglob("*"):
                filepath = filepath.resolve()
                if (
                    filepath not in self.__active_files_by_path
                ) and self.filepath_should_be_uploaded(filepath):
                    # wait until the file is actually available
                    file_ready = False
                    while not file_ready:
                        try:
                            with open(filepath) as _:
                                file_ready = True
                        except PermissionError:
                            time.sleep(0.05)
                    self.__add_active_datafile_for_path(filepath)
        except FileNotFoundError as exc:
            if retries_left > 0:
                self.__scrape_dir_for_files(retries_left=retries_left - 1)
            warnmsg = (
                f"WARNING: exhausted retries finding new files in {self.dirpath}. "
                "Check log messages below for the error encountered."
            )
            self.logger.warning(warnmsg, exc_info=exc)

    def __pull_from_handler(self):
        """
        Call the watchdog EventHandler to get paths to files that should be uploaded
        """
        to_kick_back = set()
        # get all the new files for which events have occurred
        for handler_file in self.__event_handler.get_new_files():
            new_filepath = handler_file.filepath
            # if the file isn't active yet, add it and move on
            if new_filepath not in self.__active_files_by_path:
                with self.__lock:
                    self.__add_active_datafile_for_path(new_filepath)
                continue
            # otherwise, if the file is already active in the directory
            updated_at = handler_file.last_updated
            extant_datafile = self.__active_files_by_path[new_filepath]
            # if it wasn't going to be uploaded, set it to upload
            if not extant_datafile.to_upload:
                extant_datafile.to_upload = True
            # if it's been fully produced, remove and replace it with the new version
            elif extant_datafile.fully_produced:
                with self.__lock:
                    self.__forget_inactive_files()
                    self.__add_active_datafile_for_path(new_filepath)
            # if it's been fully enqueued, or is being uploaded, send it back
            elif extant_datafile.fully_enqueued or extant_datafile.upload_in_progress:
                to_kick_back.add(handler_file)
            # if it's been waiting to upload
            elif extant_datafile.waiting_to_upload:
                # if it has been chunked before its updated time, send it back
                # otherwise nothing needs to happen
                if (
                    extant_datafile.chunked_at_timestamp
                    and extant_datafile.chunked_at_timestamp < updated_at
                ):
                    to_kick_back.add(handler_file)
            else:
                warnmsg = (
                    "WARNING: unknown status for active upload file! Will reset file to "
                    f"produce again (status message = {extant_datafile.upload_status_message})"
                )
                self.logger.warning(warnmsg)
                with self.__lock:
                    self.__forget_inactive_files()
                    self.__add_active_datafile_for_path(new_filepath)
        # send back any files that will need to be handled later
        for handler_file in to_kick_back:
            self.__event_handler.kick_back(handler_file)

    def __forget_inactive_files(self):
        """
        Read through the dictionary of active files and remove any
        that won't be uploaded or that have been fully produced
        """
        keys_to_remove = []
        for key, datafile in self.__active_files_by_path.items():
            if (not datafile.to_upload) or datafile.fully_produced:
                keys_to_remove.append(key)
        for key in keys_to_remove:
            _ = self.__active_files_by_path.pop(key)

    def __startup_from_file_registry(self):
        """
        Look through the file registry to find and prepare to enqueue any files
        that were in progress the last time the code was shut down and ignore
        any files that have previously been fully uploaded
        """
        # make sure any files listed as "in progress" have their remaining chunks enqueued
        n_files_to_resume = 0
        n_chunks_resumed = 0
        for (
            rel_filepath,
            chunks,
        ) in self.__file_registry.get_incomplete_filepaths_and_chunks():
            debugmsg = (
                f"Found {rel_filepath} in progress from a previous run. "
                f"Re-enqueuing {len(chunks)} chunks."
            )
            self.logger.debug(debugmsg)
            self.__add_chunks_for_filepath(self.dirpath / rel_filepath, chunks)
            n_files_to_resume += 1
            n_chunks_resumed += len(chunks)
        if n_files_to_resume > 0:
            infomsg = (
                f'Found {n_files_to_resume} file{"s" if n_files_to_resume!=1 else ""}'
                f" in progress from a previous run. Will re-enqueue {n_chunks_resumed} "
                f'total chunk{"s" if n_chunks_resumed!=1 else ""}'
            )
            self.logger.info(infomsg)
        # make sure any files listed as "completed" will not be uploaded again
        n_prev_uploaded = 0
        for rel_filepath in self.__file_registry.get_completed_filepaths():
            filepath = self.dirpath / rel_filepath
            debugmsg = (
                f"Found {filepath} listed as fully uploaded during a previous run. "
                "Will not produce it again."
            )
            if filepath in self.__active_files_by_path:
                if self.__active_files_by_path[filepath].to_upload:
                    self.logger.debug(debugmsg)
                self.__active_files_by_path[filepath].to_upload = False
            elif filepath.is_file():
                self.logger.debug(debugmsg)
            n_prev_uploaded += 1
        if n_prev_uploaded > 0:
            infomsg = (
                f'Found {n_prev_uploaded} file{"s" if n_prev_uploaded!=1 else ""} '
                "fully uploaded during a previous run that will not be produced again."
            )
            self.logger.info(infomsg)
        # forget any files that are now set to inactive
        self.__forget_inactive_files()

    def __add_chunks_for_filepath(self, filepath, chunks):
        """
        Add the given chunks to the list of chunks to upload for a given file
        Creates the file if it doesn't already exist in the dictionary of active files
        """
        if filepath in self.__active_files_by_path:
            self.__active_files_by_path[filepath].to_upload = True
        else:
            self.__add_active_datafile_for_path(filepath)
        self.__active_files_by_path[filepath].add_chunks_to_upload(
            chunks,
            chunk_size=self.__chunk_size,
        )

    def __restart_crashed_threads(self):
        """
        Iterate over the upload threads and restart any that crashed,
        logging the exception they threw
        """
        for ti, upload_thread in enumerate(self.__upload_threads):
            if upload_thread.caught_exception is not None:
                # log the error
                warnmsg = (
                    "WARNING: an upload thread raised an Exception, which will be logged "
                    "as an error below but not re-raised. The Producer and upload thread "
                    "that raised the error will be restarted."
                )
                self.logger.warning(warnmsg, exc_info=upload_thread.caught_exception)
                # try to join the thread
                try:
                    upload_thread.join()
                except Exception:
                    pass
                finally:
                    self.__upload_threads[ti] = None
                    self.__producers[ti] = None
                # recreate the producer and restart the thread
                self.__producers[ti] = self.get_new_producer()
                thread = ExceptionTrackingThread(
                    target=self.__producers[ti].produce_from_queue_looped,
                    args=(self.__upload_queue, self.__topic_name),
                    kwargs={"callback": self.producer_callback},
                )
                thread.start()
                self.__upload_threads[ti] = thread

    #################### CLASS METHODS ####################

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = [
            *superargs,
            "upload_dir",
            "config",
            "topic_name",
            "upload_regex",
            "chunk_size",
            "queue_max_size",
            "update_seconds",
            "upload_existing",
            "watchdog_lag_time",
        ]
        kwargs = {**superkwargs, "n_threads": RUN_CONST.N_DEFAULT_UPLOAD_THREADS}
        return args, kwargs

    @classmethod
    def run_from_command_line(cls, args=None):
        """
        Run a :class:`~DataFileUploadDirectory` directly from the command line

        Calls :func:`~upload_files_as_added` on a :class:`~DataFileUploadDirectory`
        defined by command line (or given) arguments

        :param args: the list of arguments to send to the parser
        :type args: List
        """
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        # make the DataFileDirectory for the specified directory
        upload_file_directory = cls(
            args.upload_dir,
            args.config,
            upload_regex=args.upload_regex,
            watchdog_lag_time=args.watchdog_lag_time,
            update_secs=args.update_seconds,
            streamlevel=args.logger_stream_level,
            filelevel=args.logger_file_level,
        )
        # listen for new files in the directory and run uploads until shut down
        run_start = datetime.datetime.now()
        if not args.upload_existing:
            upload_file_directory.logger.info(
                f"Listening for files to be added to {args.upload_dir}..."
            )
        else:
            upload_file_directory.logger.info(
                f"Uploading files in/added to {args.upload_dir}..."
            )
        uploaded_filepaths = upload_file_directory.upload_files_as_added(
            args.topic_name,
            n_threads=args.n_threads,
            chunk_size=args.chunk_size,
            max_queue_size=args.queue_max_size,
            upload_existing=args.upload_existing,
        )
        run_stop = datetime.datetime.now()
        upload_file_directory.logger.info(
            f"Done listening to {args.upload_dir} for files to upload"
        )
        if len(uploaded_filepaths) > 0:
            final_msg = (
                f"The following {len(uploaded_filepaths)} file"
                f'{" was" if len(uploaded_filepaths)==1 else "s were"}'
                f" uploaded between {run_start} and {run_stop}:\n\t"
            )
            final_msg += "\n\t".join([str(ufp) for ufp in uploaded_filepaths])
        else:
            final_msg = f"No files were uploaded between {run_start} and {run_stop}"
        upload_file_directory.logger.info(final_msg)

    #################### PROPERTIES ####################

    @property
    def other_datafile_kwargs(self):
        """
        Overload this property to send extra keyword arguments to the
        :class:`~UploadDataFile` constructor for each recognized file
        (useful if using a custom `datafile_type`)
        """
        return {}

    @property
    def status_msg(self):
        """
        A message stating the number of files currently at each stage of progress,
        up to the maximum number of files
        """
        self.__forget_inactive_files()
        n_wont_be_uploaded = 0
        n_waiting_to_upload = 0
        n_upload_in_progress = 0
        n_fully_enqueued = 0
        n_fully_produced = 0
        for datafile in self.__status_message_files.values():
            if not datafile.to_upload:
                n_wont_be_uploaded += 1
            elif datafile.waiting_to_upload:
                n_waiting_to_upload += 1
            elif datafile.upload_in_progress:
                n_upload_in_progress += 1
            elif datafile.fully_enqueued:
                n_fully_enqueued += 1
            elif datafile.fully_produced:
                n_fully_produced += 1
        if len(self.__status_message_files) > 0:
            status_message = (
                f"Of the {len(self.__status_message_files)} most recent files "
                f"found in {self.dirpath}, "
            )
            ns_msgs = [
                (n_wont_be_uploaded, "will not be uploaded"),
                (n_waiting_to_upload, "are waiting to upload"),
                (n_upload_in_progress, "have uploads in progress"),
                (n_fully_enqueued, "are enqueued to be produced"),
                (n_fully_produced, "are fully produced"),
            ]
            for nvar, msg in ns_msgs:
                if nvar > 0:
                    status_message += f"{nvar} {msg}, "
            status_message = f"{status_message[:-2]}"
        else:
            status_message = f"No files to be uploaded found yet in {self.dirpath}"
        return status_message

    @property
    def progress_msg(self):
        """
        A message describing the files that are currently recognized as part of the directory
        """
        self.__forget_inactive_files()
        if len(self.__status_message_files) > 0:
            msg = (
                "The following files have been recognized so far "
                f"(up to {self.N_STATUS_MESSAGE_FILES} most recent):\n\t"
            )
            msg += "\n\t".join(
                [df.upload_status_msg for df in self.__status_message_files.values()]
            )
        else:
            msg = f"No files to be uploaded found yet in {self.dirpath}"
        return msg

    @property
    def have_file_to_upload(self):
        """
        True if there are any files waiting to be uploaded or in progress
        """
        for datafile in self.__active_files_by_path.values():
            if datafile.upload_in_progress or datafile.waiting_to_upload:
                return True
        return False

    @property
    def partially_done_file_paths(self):
        """
        A list of filepaths whose uploads are currently in progress
        """
        return [
            fp for fp, df in self.__active_files_by_path.items() if df.upload_in_progress
        ]

    @property
    def n_partially_done_files(self):
        """
        The number of files currently in the process of being uploaded
        """
        return len(self.partially_done_file_paths)


def main(args=None):
    """
    Main method to run from command line
    """
    DataFileUploadDirectory.run_from_command_line(args)


if __name__ == "__main__":
    main()



"""Base class for messages computed from reconstructed data files and produced to new topics"""

# imports
from abc import ABC
from ...kafka_wrapper.producible import Producible
from ..utilities import get_message_prepend


class ReproducerMessage(Producible, ABC):
    """
    Abstract base class for messages that are computed from data files and produced to a new topic

    :param datafile: The DataFile object used to compute this message
    :type datafile: :class:`~.data_file_io.DownloadDataFileToMemory`
    """

    @property
    def msg_key(self):
        """
        The automatically defined key for these messages is a string version of the path
        to the original file with "_processing_result" appended
        """
        key_pp = get_message_prepend(self.__datafile.subdir_str, self.__datafile.filename)
        return f"{key_pp}_processing_result"

    @property
    def callback_kwargs(self):
        """
        The callback kwargs defined for these messages are the original datafile name,
        its relative filepath, and the total number of chunks into which the file was broken.
        """
        return {
            "filename": self.__datafile.filename,
            "rel_filepath": self.__datafile.relative_filepath,
            "n_total_chunks": self.__datafile.n_total_chunks,
        }

    def get_log_msg(self, print_every=None):
        """
        Prints a message saying that the processing result from [filepath] is being produced,
        if print_every is defined
        """
        if print_every:
            return f"Producing processing result from {self.__datafile.filepath}"
        return None

    def __init__(self, datafile, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__datafile = datafile



"""A single data file"""

# imports
from ...utilities import LogOwner
from ...utilities.misc import populated_kwargs


class DataFile(LogOwner):
    """
    Base class for representing a single data file

    :param filepath: the path to the file
    :type filepath: :class:`pathlib.Path`
    """

    def __init__(self, filepath, *args, **kwargs):
        self.filepath = filepath
        self.filename = self.filepath.name
        kwargs = populated_kwargs(kwargs, {"logger_file": self.filepath.parent})
        super().__init__(*args, **kwargs)



"""Example of a general Runnable class that can be installed as a Service/daemon"""

# imports
import datetime, time
from openmsistream.utilities import Runnable


class RunnableExample(Runnable):
    """
    A class that extends Runnable as an example of installing a Runnable's
    "run_from_command_line" method as a Service/daemon
    """

    @classmethod
    def get_command_line_arguments(cls):
        return ["output_dir"], {}

    @classmethod
    def run_from_command_line(cls, args=None):
        parser = cls.get_argument_parser()
        args = parser.parse_args(args)
        test_file_name = "runnable_example_service_test.txt"
        if not (args.output_dir / test_file_name).is_file():
            with open(args.output_dir / test_file_name, "w") as fp:
                msg = (
                    "This file was created to test running a generic Runnable's "
                    "'run_from_command_line' method as a Service/daemon"
                )
                fp.write(msg)
        with open(args.output_dir / test_file_name, "a") as fp:
            fp.write(
                f'\nService rerun {(datetime.datetime.now()).strftime("on %Y-%m-%d at %H:%M:%S")}'
            )
        time.sleep(30)



"""
Script to use the OS-appropriate ServceManager to install code
as a Windows Service or Linux daemon
"""

# imports
import sys
from .config import SERVICE_CONST
from .utilities import test_python_code, get_os_name
from .service_manager_base import ServiceManagerBase
from .windows_service_manager import WindowsServiceManager
from .linux_service_manager import LinuxServiceManager


def main(given_args=None):
    """
    Install a Windows Service/Linux daemon (depending on the OS that's running)
    """
    # get the arguments
    if given_args is None:
        parser = ServiceManagerBase.get_argument_parser(
            "install", sys.argv[1] if len(sys.argv) > 1 else None
        )
        args, _ = parser.parse_known_args()
    else:
        parser = ServiceManagerBase.get_argument_parser(
            "install", given_args[0] if len(given_args) > 0 else None
        )
        args, _ = parser.parse_known_args(given_args)
    # run the tests if requested
    if args.service_spec_string == "test":
        test_python_code()
    else:
        if args.service_name is None:
            # default name of the Service is just the class name
            service_name = args.service_spec_string
        else:
            service_name = args.service_name
        # make the list of arguments that should be sent to the run function in the executable
        argslist = sys.argv[2:] if given_args is None else given_args[1:]
        if "--service_name" in argslist:
            index = argslist.index("--service_name")
            argslist.pop(index)
            argslist.pop(index)
        # define arguments to and create the ServiceManager
        manager_args = [service_name]
        manager_kwargs = {
            "service_spec_string": args.service_spec_string,
            "argslist": argslist,
            "interactive": given_args is None,
            "logger": SERVICE_CONST.logger,
        }
        managers_by_os_name = {
            "Windows": WindowsServiceManager,
            "Linux": LinuxServiceManager,
        }
        manager_class = managers_by_os_name[get_os_name()]
        manager = manager_class(*manager_args, **manager_kwargs)
        # install the service
        manager.install_service()


if __name__ == "__main__":
    main()



""" A watchdog FileSystemEventHandler for use with a DataFileUploadDirectory """

# imports
import pathlib, datetime
from dataclasses import dataclass
from threading import RLock
from watchdog.events import FileSystemEventHandler
from ...utilities import LogOwner


@dataclass
class EventHandlerActiveFile:
    """
    A small dataclass to hold active files recognized by the event handler,
    along with a timestamp
    """

    filepath: pathlib.Path
    last_updated: datetime.datetime


class UploadDirectoryEventHandler(LogOwner, FileSystemEventHandler):
    """
    An extension of a watchdog FileSystemEventHandler to handle file events used by
    :class:`~DataFileUploadDirectory` objects. Keeps a dictionary of files with
    timestamps of when they were updated. That dictionary can be polled for active
    file objects by the upload directory.

    :param upload_regex: only files matching this regular expression will be uploaded
    :type upload_regex: :func:`re.compile`, optional
    :param logs_subdir: path to the owning upload directory's "logs" directory
    :type logs_subdir: :class:`pathlib.Path`
    :param watchdog_lag_time: Number of seconds that files must remain static (unmodified)
        before it's given in :func:`get_new_files`
    :type watchdog_lag_time: int
    """

    #################### NEW PUBLIC FUNCTIONS ####################

    def __init__(self, upload_regex, logs_subdir, lag_time, **other_kwargs):
        """
        Constructor method
        """
        super().__init__(**other_kwargs)
        # variables for determining files whose events should be dispatched
        self.__upload_regex = upload_regex
        self._logs_subdir = logs_subdir
        self.__rootdir = self._logs_subdir.parent
        self.__lag_time = lag_time
        # A thread lock
        self.__lock = RLock()
        # all currently active files
        self.__active_files_by_path = {}

    def get_new_files(self):
        """
        A generator of any new EventHandlerActiveFile objects that have persisted
        for at least a small amount of time. Called by the upload directory to get
        new files to upload (or re-upload modified or moved files)
        """
        keys_to_pop = []
        ref_timestamp = datetime.datetime.now()
        with self.__lock:
            for filepath, active_file in self.__active_files_by_path.items():
                if (
                    ref_timestamp - active_file.last_updated
                ).total_seconds() > self.__lag_time:
                    keys_to_pop.append(filepath)
            for key in keys_to_pop:
                yield self.__active_files_by_path.pop(key)

    def kick_back(self, handler_file):
        """
        Give back an :class:`EventHandlerActiveFile` so it can be processed later

        :param handler_file: The file to add back in as active
        :type handler_file: :class:`EventHandlerActiveFile`
        """
        self.__add_active_file(handler_file)

    def filepath_matched(self, filepath):
        """
        Returns True if a given filepath should be considered for event dispatch
        (or, eventually, upload). Only returns True for paths to files relative
        to the upload directory root dir, but not in the logs subdir, with names
        that don't start with "." and whose relative paths match the upload regex

        :param filepath: Filepath to test
        :type filepath: :class:`pathlib.Path`

        :raises TypeError: if `filepath` is not a path
        """
        if not isinstance(filepath, pathlib.PurePath):
            self.logger.error(
                f"ERROR: {filepath} passed to filepath_should_be_uploaded is not a Path!",
                exc_type=TypeError,
            )
        # only files count
        if not filepath.is_file():
            return False
        # must be relative to the upload directory
        try:
            if not filepath.is_relative_to(self.__rootdir):
                return False
        except AttributeError:  # "is_relative_to" was added after 3.7
            if not str(filepath).startswith(str(self.__rootdir)):
                return False
        # must be outside the logs subdirectory
        if self._logs_subdir in filepath.parents:
            return False
        # name shouldn't start with '.'
        if filepath.name.startswith("."):
            return False
        # must match the given regex
        if self.__upload_regex.match(str(filepath.relative_to(self.__rootdir))):
            return True
        return False

    #################### INHERITED FUNCTIONS ####################

    def dispatch(self, event):
        # Ignore directory events
        if event.is_directory:
            return
        # add the source and destination paths
        paths = []
        if event.src_path:
            paths.append(pathlib.Path(event.src_path))
        if hasattr(event, "dest_path"):
            paths.append(pathlib.Path(event.dest_path))
        # only dispatch events with at least one of their paths being relevant
        if any(self.filepath_matched(p) for p in paths):
            super().dispatch(event)

    def on_moved(self, event):
        super().on_moved(event)
        src_path = pathlib.Path(event.src_path)
        dest_path = pathlib.Path(event.dest_path)
        with self.__lock:
            self.__remove_active_file(src_path)
            self.__add_active_file(dest_path)

    def on_created(self, event):
        super().on_created(event)
        src_path = pathlib.Path(event.src_path)
        self.__add_active_file(src_path)

    def on_deleted(self, event):
        super().on_deleted(event)
        src_path = pathlib.Path(event.src_path)
        self.__remove_active_file(src_path)

    def on_modified(self, event):
        super().on_modified(event)
        src_path = pathlib.Path(event.src_path)
        self.__add_active_file(src_path)

    #################### PRIVATE HELPER FUNCTIONS ####################

    def __add_active_file(self, filepath_or_handler_file):
        if isinstance(filepath_or_handler_file, pathlib.PurePath):
            with self.__lock:
                self.__active_files_by_path[
                    filepath_or_handler_file
                ] = EventHandlerActiveFile(
                    filepath_or_handler_file, datetime.datetime.now()
                )
        elif isinstance(filepath_or_handler_file, EventHandlerActiveFile):
            with self.__lock:
                self.__active_files_by_path[
                    filepath_or_handler_file.filepath
                ] = filepath_or_handler_file
        else:
            self.logger.error(
                f"ERROR: unrecognized type {type(filepath_or_handler_file)}",
                exc_type=TypeError,
            )

    def __remove_active_file(self, filepath):
        with self.__lock:
            if filepath in self.__active_files_by_path:
                _ = self.__active_files_by_path.pop(filepath)



"""The base message type that must be produced by a MetadataJSONReproducer"""

# imports
import json
from ..data_file_io.entity.reproducer_message import ReproducerMessage


class MetadataJSONMessage(ReproducerMessage):
    """
    A class to represent the json metadata from a file as a Producible
    """

    @property
    def msg_value(self):
        if self.json_content is None:
            raise ValueError(
                f"ERROR: JSON content should not be None for a {self.__class__.__name__}!"
            )
        return json.dumps(self.json_content)

    def __init__(self, *args, json_content=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.json_content = json_content



# imports
from io import BytesIO
from hashlib import sha256
import girder_client
from ..version import __version__
from ..data_file_io.actor.data_file_stream_processor import DataFileStreamProcessor
from ..utilities.config import RUN_CONST


class GirderUploadStreamProcessor(DataFileStreamProcessor):
    """
    A class to reconstruct data files read as messages from a topic, hold them in memory
    or on disk, and upload them to a Girder instance when all of their messages have been
    received

    :param girder_api_url: URL of the REST API endpoint for the Girder instance to which
        files should be uploaded
    :type girder_api_url: str
    :param girder_api_key: API key for interacting with the Girder instance
    :type girder_api_key: str
    :param config_file: Path to the config file to use in defining the Broker connection
        and Consumers
    :type config_file: :class:`pathlib.Path`
    :param topic_name: Name of the topic to which the Consumers should be subscribed
    :type topic_name: str
    :param girder_root_folder_ID: ID of an existing Girder Folder relative to which
        files should be uploaded. Additional Folders will be created within this root
        Folder to replicate the original Producer-side subdirectory structure.
    :type girder_root_folder_ID: str
    :param collection_name: Name of the Girder Collection to which files should be
        uploaded. Only used if `girder_root_folder_id` is not given.
    :type collection_name: str
    :param girder_root_folder_path: Path to the root Folder within which files should be
        uploaded. Additional Folders will be created within this root Folder to replicate
        the original Producer-side subdirectory structure. This argument is only used if
        `girder_root_folder_id` is not provided. If a `collection_name` is given but this
        argument is not, a Folder named after the topic will be created within the
        Collection.
    :type girder_root_folder_path: str
    :param provider: If this argument is given, an extra "provider" metadata field with
        the given value will be added to uploaded Files and Folders.
    :type provider: str
    :param filepath_regex: If given, only messages associated with files whose paths match
        this regex will be consumed
    :type filepath_regex: :type filepath_regex: :func:`re.compile` or None, optional
    """

    def __init__(
        self,
        girder_api_url,
        girder_api_key,
        config_file,
        topic_name,
        *,
        girder_root_folder_id=None,
        collection_name=None,
        girder_root_folder_path=None,
        provider=None,
        **other_kwargs,
    ):
        super().__init__(config_file, topic_name, **other_kwargs)
        # connect and authenticate to the Girder instance
        try:
            self.__girder_client = girder_client.GirderClient(apiUrl=girder_api_url)
            self.__girder_client.authenticate(apiKey=girder_api_key)
        except Exception as exc:
            errmsg = (
                f"ERROR: failed to authenticate to the Girder instance at {girder_api_url}. "
                "Exception will be re-raised."
            )
            self.logger.error(errmsg, exc_info=exc, reraise=True)
        # set some minimal amount of metadata fields
        self.minimal_metadata_dict = {
            "OpenMSIStreamVersion": __version__,
            "KafkaTopic": topic_name,
        }
        if provider:
            self.minimal_metadata_dict["provider"] = provider
        # if a root folder ID was given, just use that
        if girder_root_folder_id:
            self.__root_folder_id = girder_root_folder_id
        # otherwise, figure it out from the given collection name and/or root folder path
        else:
            # get or create the collection with the given name, and save its ID
            if not collection_name:
                collection_name = RUN_CONST.DEFAULT_COLLECTION_NAME
            collection_id = self.__init_collection(collection_name)
            # get or create the root folder and save its ID
            root_folder_path = girder_root_folder_path
            if not root_folder_path:
                root_folder_path = f"{collection_name}/{topic_name}"
            self.__root_folder_id = self.__init_root_folder(
                root_folder_path, collection_id, collection_name
            )

    def _process_downloaded_data_file(self, datafile, lock):
        """
        Upload a fully-reconstructed file to the Girder instance, creating Folders as
        necessary to preserve original subdirectory structure. Also adds metadata to Files
        and Folders listing the version of OpenMSIStream that's running and the name of
        the topic from which files are being consumed.

        :param datafile: A :class:`~DownloadDataFile` object that has received
            all of its messages from the topic
        :type datafile: :class:`~DownloadDataFile`
        :param lock: Acquiring this :class:`threading.Lock` object ensures that only one instance
            of :func:`~_process_downloaded_data_file` is running at once
        :type lock: :class:`threading.Lock`

        :return: None if upload was successful, a caught Exception otherwise
        """
        # Create the nested subdirectories that should hold this file
        parent_id = self.__root_folder_id
        if datafile.subdir_str != "":
            subdir_str_split = datafile.subdir_str.split("/")
            for folder_name in subdir_str_split:
                metadata_dict = self.minimal_metadata_dict.copy()
                try:
                    new_folder_id = self.__create_folder(
                        parent_id,
                        folder_name,
                        parentType="folder",
                        reuseExisting=True,
                    )
                    self.__girder_client.addMetadataToFolder(
                        new_folder_id,
                        metadata_dict,
                    )
                except Exception as exc:
                    errmsg = f"ERROR: failed to create the '{folder_name}' folder"
                    self.logger.error(errmsg, exc_info=exc)
                    return exc
                parent_id = new_folder_id
        else:
            subdir_str_split = []
        # Upload the file from its bytestring or file on disk
        try:
            with lock:
                try:
                    self.__girder_client.uploadStreamToFolder(
                        parent_id,
                        BytesIO(datafile.bytestring),
                        datafile.filename,
                        len(datafile.bytestring),
                    )
                except AttributeError:
                    self.__girder_client.uploadFileToFolder(
                        parent_id, datafile.full_filepath
                    )
        except Exception as exc:
            errmsg = f"ERROR: failed to upload the file at {datafile.relative_filepath}"
            self.logger.error(errmsg, exc_info=exc)
            return exc
        # Add metadata to the item that was created for the file
        item_id = None
        for resp in self.__girder_client.listItem(parent_id, name=datafile.filename):
            if item_id:
                errmsg = (
                    f"ERROR: found more than one Item named {datafile.filename} "
                    f"after uploading the file at {datafile.relative_filepath}"
                )
                return RuntimeError(errmsg)
            item_id = resp["_id"]
        if not item_id:
            errmsg = (
                "ERROR: could not find a corresponding Item after uploading the file "
                f"at {datafile.relative_filepath}"
            )
            return RuntimeError(errmsg)
        checksum_hash = sha256()
        checksum_hash.update(datafile.bytestring)
        metadata_dict = self.minimal_metadata_dict.copy()
        metadata_dict["checksum"] = {
            "sha256": checksum_hash.hexdigest(),
        }
        try:
            self.__girder_client.addMetadataToItem(item_id, metadata_dict)
        except Exception as exc:
            errmsg = (
                "ERROR: failed to set metadata for the Item corresponding to the file "
                f"uploaded from {datafile.relative_filepath}"
            )
            self.logger.error(errmsg, exc_info=exc)
            return exc
        return None

    def __init_collection(self, collection_name):
        """
        Find or create a Collection with the given name, returning its ID

        Logs and re-raises any Exceptions encountered

        Returns the ID of the Collection with the given name
        """
        collection_id = None
        try:
            for resp in self.__girder_client.listCollection():
                if resp["_modelType"] == "collection" and resp["name"] == collection_name:
                    collection_id = resp["_id"]
            if not collection_id:
                new_collection = self.__girder_client.createCollection(
                    collection_name, public=True
                )
                collection_id = new_collection["_id"]
        except Exception as exc:
            errmsg = (
                f"ERROR: failed to find or create a collection called {collection_name}. "
                "Exception will be re-raised."
            )
            self.logger.error(errmsg, exc_info=exc, reraise=True)
        return collection_id

    def __init_root_folder(self, root_folder, collection_id, collection_name):
        """
        Find or create a "root" Girder Folder at the given (posix-formatted string) path
        into which files should be reconstructed

        Returns the ID of the root folder and its relative path string for metadata use
        """
        root_folder_split = root_folder.split("/")
        start_index = 1 if root_folder_split[0] == collection_name else 0
        parent_folder_id = None
        for folder_depth, folder_name in enumerate(root_folder_split):
            metadata_dict = self.minimal_metadata_dict.copy()
            new_folder_id = self.__create_folder(
                parent_folder_id if parent_folder_id else collection_id,
                folder_name,
                parentType="folder" if parent_folder_id else "collection",
                reuseExisting=True,
                metadata=metadata_dict if folder_depth >= start_index else None,
            )
            parent_folder_id = new_folder_id
        root_folder_id = parent_folder_id
        return root_folder_id

    def __create_folder(self, parent_id, name, **kwargs):
        """
        Create a new folder in the Girder instance under the given parent ID
        with the given name

        Keyword arguments are passed to GirderClient.createFolder

        Logs and re-raises any encountered exceptions

        Returns the ID of the created folder
        """
        new_folder_id = None
        try:
            new_folder = self.__girder_client.createFolder(
                parent_id,
                name,
                **kwargs,
            )
            new_folder_id = new_folder["_id"]
        except Exception as exc:
            errmsg = (
                f"ERROR: failed to find or create a folder called {name}. "
                "Exception will be re-raised."
            )
            self.logger.error(errmsg, exc_info=exc, reraise=True)
        return new_folder_id

    @classmethod
    def get_command_line_arguments(cls):
        """
        Return the names of arguments needed to run the program from the command line
        """
        superargs, superkwargs = super().get_command_line_arguments()
        ret_args = [
            "girder_api_url",
            "girder_api_key",
            *superargs,
            "girder_root_folder_id",
            "collection_name",
            "girder_root_folder_path",
            "provider",
        ]
        return ret_args, superkwargs

    @classmethod
    def run_from_command_line(cls, args=None):
        """
        Run a :class:`~GirderUploadStreamProcessor` directly from the command line

        Calls :func:`~DataFileStreamProcessor.process_files_as_read` on a
        :class:`~GirderUploadStreamProcessor` defined by command line (or given) arguments

        :param args: the list of arguments to send to the parser instead of getting them
            from sys.argv
        :type args: list, optional
        """
        # make the argument parser
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        # make the stream processor
        girder_uploader = cls(
            args.girder_api_url,
            args.girder_api_key,
            args.config,
            args.topic_name,
            girder_root_folder_id=args.girder_root_folder_id,
            collection_name=args.collection_name,
            girder_root_folder=args.girder_root_folder,
            provider=args.provider,
            output_dir=args.output_dir,
            mode=args.mode,
            n_threads=args.n_threads,
            update_secs=args.update_seconds,
            consumer_group_id=args.consumer_group_id,
        )
        # start the processor running
        msg = (
            f"Listening to the {args.topic_name} topic for files to upload to "
            f"Girder using the API at {args.girder_api_url}"
        )
        girder_uploader.logger.info(msg)
        (
            n_read,
            n_msgs_procd,
            n_files_procd,
            procd_fps,
        ) = girder_uploader.process_files_as_read()
        # shut down when that function returns
        girder_uploader.close()
        msg = "Girder upload stream processor "
        if args.output_dir is not None:
            msg += f"writing to {args.output_dir} "
        msg += "shut down"
        girder_uploader.logger.info(msg)
        msg = (
            f"{n_read} total messages were consumed, {n_msgs_procd} messages were "
            f"successfully processed, and {n_files_procd} files were uploaded "
            f"to Girder"
        )
        girder_uploader.logger.info(msg)
        if len(procd_fps) > 0:
            msg = (
                f"{n_files_procd} file{' was' if n_files_procd==1 else 's were'} "
                f"successfully uploaded to Girder."
                f"\nUploaded filepaths (up to {cls.N_RECENT_FILES} most recent):\n\t"
            )
            msg += "\n\t".join([str(fp) for fp in procd_fps])
            girder_uploader.logger.debug(msg)


def main(args=None):
    """
    Run the stream processor from the command line
    """
    GirderUploadStreamProcessor.run_from_command_line(args)


if __name__ == "__main__":
    main()



"""Expands a ConfigFileParser for the S3 section of configs"""

# imports
from ..utilities.config_file_parser import ConfigFileParser


class S3ConfigFileParser(ConfigFileParser):
    """
    A class to parse S3 bucket configurations from files
    """

    @property
    def s3_configs(self):
        """
        The S3 section of the config file
        """
        if self.__s3_configs is None:
            self.__s3_configs = self._get_config_dict("s3")
        return self.__s3_configs

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__s3_configs = None



# imports
import setuptools, pathlib

# read version tag
version = None
version_path = pathlib.Path(__file__).parent/"openmsistream"/"version.py"
with open(version_path,"r") as version_file:
    for line in version_file.readlines():
        if line.startswith("__version__"):
            version = line.strip().split("=")[-1].strip().strip('"')
if not version:
    raise RuntimeError("ERROR: Failed to find version tag!")

# read a portion of the README to get the description
long_description = ""
with open("README.md", "r") as readme:
    for il, line in enumerate(readme.readlines(), start=1):
        if il >= 18:
            long_description += line

setupkwargs = dict(
    name="openmsistream",
    packages=setuptools.find_packages(include=["openmsistream*"]),
    include_package_data=True,
    version=version,
    description=(
        "Python applications for materials data streaming using Apache Kafka. "
        "Developed for Open MSI (NSF DMREF award #1921959)"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="OpenMSIStream",
    author_email="openmsistream@gmail.com",
    url="https://github.com/openmsi/openmsistream",
    download_url=f"https://github.com/openmsi/openmsistream/archive/refs/tags/v{version}.tar.gz",
    license="GNU GPLv3",
    entry_points={
        "console_scripts": [
            "UploadDataFile=openmsistream.data_file_io.entity.upload_data_file:main",
            "DataFileUploadDirectory=openmsistream.data_file_io.actor.data_file_upload_directory:main",
            "DataFileDownloadDirectory=openmsistream.data_file_io.actor.data_file_download_directory:main",
            "InstallService=openmsistream.services.install_service:main",
            "ManageService=openmsistream.services.manage_service:main",
            "ProvisionNode=openmsistream.utilities.provision_wrapper:main",
            "S3TransferStreamProcessor=openmsistream.s3_buckets.s3_transfer_stream_processor:main",
            "GirderUploadStreamProcessor=openmsistream.girder.girder_upload_stream_processor:main",
        ],
    },
    python_requires=">=3.7,<3.10",
    install_requires=[
        "atomicwrites>=1.4.1",
        "boto3>=1.26.84",
        "confluent-kafka>=2.0.2",
        "girder-client>=3.1.20",
        "kafkacrypto>=0.9.10.0",
        "matplotlib",
        "methodtools",
        "msgpack",
        "watchdog>=3.0.0",
    ],
    extras_require={
        "test": [
            "beautifulsoup4",
            "black",
            "docker",
            "gitpython",
            "lxml",
            "marko[toc]==1.3.0",
            "pyflakes>=3.0.1",
            "pylint>=2.16.3",
            "requests",
            "tempenv>=2.0.0",
        ],
        "docs": [
            "sphinx>=6.1.3",
            "sphinx_rtd_theme>=1.2.0",
        ],
        "dev": [
            "twine",
        ],
    },
    keywords=[
        "data_streaming",
        "stream_processing",
        "apache_kafka",
        "materials",
        "data_science",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)

setupkwargs["extras_require"]["all"] = sum(setupkwargs["extras_require"].values(), [])

setuptools.setup(**setupkwargs)



"""A Thread that re-raises any Exceptions it encounters during running when it is join()ed"""

# imports
from threading import Thread


class ExceptionTrackingThread(Thread):
    """
    A small utility class to keep track of any exceptions thrown in a child thread
    and raise them at some point
    """

    @property
    def caught_exception(self):
        """
        Variable to hold any Exception encountered while the thread is running
        """
        return self.__exc

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__exc = None

    def run(self, *args, **kwargs):
        """
        Wrapper around Thread.run that holds onto any Exception raised during running
        """
        try:
            super().run(*args, **kwargs)
        except Exception as exc:
            self.__exc = exc

    def join(self, *args, **kwargs):
        """
        Wrapper around Thread.run that re-raises any Exceptions that were encountered
        """
        super().join(*args, **kwargs)
        if self.__exc is not None:
            raise self.__exc



"""OpenMSIStream library for simplified data streaming using Apache Kafka"""

# imports
import os
from .data_file_io.entity.upload_data_file import UploadDataFile
from .data_file_io.actor.data_file_upload_directory import DataFileUploadDirectory
from .data_file_io.actor.data_file_download_directory import DataFileDownloadDirectory
from .data_file_io.actor.data_file_stream_processor import DataFileStreamProcessor
from .data_file_io.actor.data_file_stream_reproducer import DataFileStreamReproducer
from .s3_buckets.s3_transfer_stream_processor import S3TransferStreamProcessor
from .girder.girder_upload_stream_processor import GirderUploadStreamProcessor
from .version import __version__

__all__ = [
    "__version__",
    "UploadDataFile",
    "DataFileUploadDirectory",
    "DataFileDownloadDirectory",
    "DataFileStreamProcessor",
    "DataFileStreamReproducer",
    "S3TransferStreamProcessor",
    "GirderUploadStreamProcessor",
]

# If running on Windows, try to pre-load the librdkafka.dll file
if os.name == "nt":
    try:
        import confluent_kafka
    except Exception:
        import traceback

        reason = None
        import sys, pathlib

        dll_dir = (
            pathlib.Path(sys.executable).parent
            / "Lib"
            / "site-packages"
            / "confluent_kafka.libs"
        )
        if not dll_dir.is_dir():
            reason = f"expected DLL directory {dll_dir} does not exist!"
        from ctypes import CDLL

        fps = []
        for fp in dll_dir.glob("*.dll"):
            if fp.name.startswith("librdkafka"):
                fps.append(fp)
                CDLL(str(fp))
        if len(fps) < 1:
            reason = f"{dll_dir} does not contain any librdkafka DLL files to preload!"
        if reason is not None:
            print(f"WARNING: Failed to preload librdkafka DLLs. Reason: {reason}")
        try:
            import confluent_kafka
        except Exception as exc:
            errmsg = "ERROR: Preloading librdkafka DLLs ("
            errmsg += ", ".join([str(_) for _ in fps])
            errmsg += (
                f"{errmsg}) did not allow confluent_kafka to be imported! "
                f"Exception (will be re-raised): {traceback.format_exc()}"
            )
            print(errmsg)
            raise exc
    _ = confluent_kafka.Producer  # appease pyflakes



"""
A controlled process that is only running in a single thread
"""

# imports
from abc import ABC, abstractmethod
from .controlled_process import ControlledProcess


class ControlledProcessSingleThread(ControlledProcess, ABC):
    """
    A class for running a process in a single thread in a loop until it's explicitly shut down.
    Every iteration a single function is called, and then the control command queue is checked.
    """

    def run(self):
        """
        Start the process and call :func:`~_run_iteration` until the process is shut down
        """
        super().run()
        while self.alive:
            self._run_iteration()
            self._print_still_alive()
            self._check_control_command_queue()

    @abstractmethod
    def _run_iteration(self):
        """
        The function that is called repeatedly in an infinite loop as long as the process is alive.

        Not implemented in the base class.
        """
        raise NotImplementedError



"""
A ControlledProcess running with more than one thread
"""

# imports
from threading import Lock
from abc import ABC, abstractmethod
from ..utilities.exception_tracking_thread import ExceptionTrackingThread
from .config import RUN_CONST
from .controlled_process import ControlledProcess


class ControlledProcessMultiThreaded(ControlledProcess, ABC):
    """
    A class for running a group of processes in multiple threads until they're explicitly shut down.

    :param n_threads: the number of threads to use in the group of processes
    :type n_threads: int, optional
    """

    # time in seconds to wait for threads to join in shutdown
    SHUTDOWN_THREAD_TIMEOUT = 10

    def __init__(self, *args, n_threads=RUN_CONST.DEFAULT_N_THREADS, **kwargs):
        self.n_threads = n_threads
        super().__init__(*args, **kwargs)
        self.lock = Lock()
        self.__args_per_thread = []
        self.__kwargs_per_thread = {}
        self.__threads = []

    def run(self, args_per_thread=None, kwargs_per_thread=None):
        """
        Create and start the set of independent threads running :func:`~_run_worker`,
        which can accept optional arguments/keyword arguments from the call to this function.
        Once the threads are started they will be automatically restarted if they crash.

        :param args_per_thread: a list of lists of arguments that should be given to the
            independent threads (one list of arguments per thread)
        :type args_per_thread: list, optional
        :param kwargs_per_thread: a list of dicts of keyword arguments that should be given
            to the independent threads (one dict of keyword arguments per thread)
        :type kwargs_per_thread: dict, optional
        """
        super().run()
        # correct the arguments for each thread
        if args_per_thread is not None:
            self.__args_per_thread = args_per_thread
        if self.__args_per_thread == [] or (not isinstance(self.__args_per_thread, list)):
            self.__args_per_thread = [self.__args_per_thread]
        if not len(self.__args_per_thread) == self.n_threads:
            if not len(self.__args_per_thread) == 1:
                errmsg = (
                    f"ERROR: ControlledProcessMultiThreaded.run was given a list of arguments "
                    f"with {len(self.__args_per_thread)} entries, but was set up to use "
                    f"{self.n_threads} threads!"
                )
                self.logger.error(errmsg, exc_type=ValueError)
            else:
                self.__args_per_thread = self.n_threads * self.__args_per_thread
        # correct the keyword arguments for each thread
        if kwargs_per_thread is not None:
            self.__kwargs_per_thread = kwargs_per_thread
        if not isinstance(self.__kwargs_per_thread, list):
            self.__kwargs_per_thread = [self.__kwargs_per_thread]
        if not len(self.__kwargs_per_thread) == self.n_threads:
            if not len(self.__kwargs_per_thread) == 1:
                errmsg = (
                    f"ERROR: ControlledProcessMultiThreaded.run was given a list of arguments "
                    f"with {len(self.__kwargs_per_thread)} entries, but was set up to use "
                    f"{self.n_threads} threads!"
                )
                self.logger.error(errmsg, exc_type=ValueError)
            else:
                self.__kwargs_per_thread = self.n_threads * self.__kwargs_per_thread
        # create and start the independent threads
        for i in range(self.n_threads):
            self.__threads.append(
                ExceptionTrackingThread(
                    target=self._run_worker,
                    args=self.__args_per_thread[i],
                    kwargs=self.__kwargs_per_thread[i],
                )
            )
            self.__threads[-1].start()
        # loop while the process is alive
        while self.alive:
            self._print_still_alive()
            self._check_control_command_queue()
            self.__restart_crashed_threads()

    def _on_shutdown(self):
        """
        Join all of the running threads. Can override this method further in subclasses,
        just be sure to also call :func:`super()._on_shutdown()`.
        """
        for thread in self.__threads:
            thread.join(self.SHUTDOWN_THREAD_TIMEOUT)

    @abstractmethod
    def _run_worker(self):
        """
        The function that will actually be run in multiple threads while the process is alive.
        Should include a "while self.alive" loop to keep it running.

        `self.lock` can be used to guarantee exactly one copy of this function is running at a time.

        Not implemented in the base class.
        """
        raise NotImplementedError

    def __restart_crashed_threads(self):
        """
        Log (but don't re-raise) Exceptions thrown by any of the threads and restart them
        based on the args/kwargs initially passed to :func:`~run`.
        """
        for ti, thread in enumerate(self.__threads):
            if thread.caught_exception is not None:
                # log the error
                warnmsg = (
                    "WARNING: a thread raised an Exception, which will be logged as an "
                    "error below but not re-raised. The thread that raised the error "
                    "will be restarted."
                )
                self.logger.warning(warnmsg, exc_info=thread.caught_exception)
                # try to join the thread
                try:
                    thread.join()
                except Exception:
                    pass
                finally:
                    self.__threads[ti] = None
                # restart the thread
                self.__threads[ti] = ExceptionTrackingThread(
                    target=self._run_worker,
                    args=self.__args_per_thread[ti],
                    kwargs=self.__kwargs_per_thread[ti],
                )
                self.__threads[ti].start()



"""Some miscellaneous functions that are universally internally available"""

# imports
import os, sys, inspect, time, contextlib


def add_user_input(input_queue):
    """
    Listen for and add user input to a queue at one second intervals
    """
    while True:
        time.sleep(1)
        input_queue.put((sys.stdin.read(1)).strip())


def raise_err_with_optional_logger(logger, errmsg, exc_type):
    """
    If logger is not None, log the error message with the given exception type,
    otherwise raise the error
    """
    if logger is not None:
        logger.error(errmsg, exc_type=exc_type)
    else:
        raise exc_type(errmsg)


def debug_msg_with_optional_logger(logger, msg):
    """
    Log a given message at debug level if logger is not None, otherwise print it
    """
    if logger is not None:
        logger.debug(msg)
    else:
        print(msg)


def _ensure_classes_or_types_match_for_key(key, test, options, logger=None):
    """
    Raise errors if "test" is not a subclass of the possibilities in "options"
    if "options" holds classes, or if the type of "test" is not in "options"
    if "options" holds objects
    """
    # if the options are classes
    if inspect.isclass(options[0] if isinstance(options, tuple) else options):
        # make sure the test argument is also a class
        if not inspect.isclass(test):
            errmsg = f'ERROR: Argument "{key}" expected to be a class, not an object!'
            raise_err_with_optional_logger(logger, errmsg, RuntimeError)
        # make sure the test argument is a subtype of any of the classes in the defaults
        if not issubclass(test, options[1:] if isinstance(options, tuple) else options):
            errmsg = (
                f'ERROR: Class type mismatch for argument "{key}", got {test} '
                f"but expected {options[1:] if isinstance(options,tuple) else options}"
            )
            raise_err_with_optional_logger(logger, errmsg, RuntimeError)
    # if the options are objects instead, make sure the type of the test argument matches
    # one of the possible types in the defaults
    elif (isinstance(options, tuple) and type(test) not in options[1:]) or (
        not isinstance(options, tuple) and not isinstance(test, type(options))
    ):
        errmsg = f'ERROR: Type mismatch replacing argument "{key}" with {test} (expected '
        if isinstance(options, tuple):
            for typestring in options[1:]:
                errmsg += f"{typestring}, "
        else:
            errmsg += f"{type(options)}"
        errmsg += f"but got {type(test)})"
        raise_err_with_optional_logger(logger, errmsg, TypeError)


def populated_kwargs(given_kwargs, defaults, logger=None):
    """
    Return a kwargs dictionary where every possible entry from the defaults has a valid value
    Can use this to make sure certain entries are present in kwargs
    """
    # for all the needed keys
    for key in defaults.keys():
        # if it was given and is not None
        if key in given_kwargs.keys() and given_kwargs[key] is not None:
            # if there's just the default option it doesn't need to be a tuple
            if isinstance(defaults[key], tuple):
                if len(defaults[key]) == 1:
                    defaults[key] = (defaults[key])[0]
            # check the options for what the argument is allowed to be
            _ensure_classes_or_types_match_for_key(
                key, given_kwargs[key], defaults[key], logger
            )
        # if it wasn't given, then just add it to the dictionary as the default
        else:
            if isinstance(defaults[key], tuple):
                given_kwargs[key] = (defaults[key])[0]
            else:
                given_kwargs[key] = defaults[key]
    return given_kwargs


@contextlib.contextmanager
def change_dir(dirpath):
    """
    Change the current working directory to a different directory,
    and go back when leaving the context manager.
    """
    cdminus = os.getcwd()
    try:
        yield os.chdir(dirpath)
    finally:
        os.chdir(cdminus)



"""A wrapped Kafka Consumer. May consume encrypted messages."""

# imports
import uuid, warnings, methodtools
from confluent_kafka import DeserializingConsumer, Message

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto import KafkaConsumer
from ..utilities import LogOwner
from ..utilities.misc import (
    raise_err_with_optional_logger,
    debug_msg_with_optional_logger,
)
from ..data_file_io.entity.data_file_chunk import DataFileChunk
from .utilities import add_kwargs_to_configs, KCCommitOffsetDictKey, KCCommitOffset
from .config_file_parser import KafkaConfigFileParser
from .openmsistream_kafka_crypto import OpenMSIStreamKafkaCrypto
from .serialization import CompoundDeserializer


class OpenMSIStreamConsumer(LogOwner):
    """
    Wrapper for working with a Consumer of some type. Expects message values that are
    :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk` objects by default;
    other message value types can be accommodated by setting "value.deserializer"
    in the config file.

    :param consumer_type: The type of underlying Consumer that should be used
    :type consumer_type: :class:`confluent_kafka.DeserializingConsumer` or
        :class:`kafkacrypto.KafkaConsumer`
    :param configs: A dictionary of configuration names and parameters to use in instantiating
        the underlying Consumer
    :type configs: dict
    :param kafkacrypto: The :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto` object
        that should be used to instantiate the Consumer. Only needed if `consumer_type` is
        :class:`kafkacrypto.KafkaConsumer`.
    :type kafkacrypto: :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto`, optional
    :param message_key_regex: A regular expression to filter messages based on their keys.
        Only messages matching this regex will be returned by :func:`~get_next_message`.
        This parameter only has an effect if `restart_at_beginning` is true and a consumer group
        with the given ID has previously consumed messages from the topic, or if
        `filter_new_message_keys` is True. Messages with keys that are not strings will always
        be consumed, logging warnings if they should be filtered.
    :type message_key_regex: :func:`re.compile` or None, optional
    :param filter_new_message_keys: If False (the default) the `message_key_regex` will only be used
        to filter messages from before each partition's currently committed location.
        Useful if you want to process only some messages from earlier in the topic, but all new
        messages that have never been read. To filter every message read instead of just
        previously-consumed messages, set this to True.
    :type filter_new_message_keys: bool, optional
    :param starting_offsets: A list of :class:`confluent_kafka.TopicPartition` objects listing
        the initial starting offsets for each partition in the topic for Consumers with this
        group ID. If `filter_new_message_keys` is False, messages with offsets greater than or
        equal to these will NOT be filtered using `message_key_regex`.
    :type starting_offsets: list(:class:`confluent_kafka.TopicPartition`) or None, optional
    :param filepath_regex: A regular expression to filter messages based on the path to the
        file with which they're associated (relative to the "root" upload directory).
        Only messages matching this regex will be returned by :func:`~get_next_message`.
        Messages that cannot be deserialized to :class:`~DataFileChunk` objects will
        always be consumed, logging warnings if they should be filtered.
    :type filepath_regex: :func:`re.compile` or None, optional
    :param kwargs: Any extra keyword arguments are added to the configuration dict for the Consumer,
        with underscores in their names replaced by dots
    :type kwargs: dict

    :raises ValueError: if `consumer_type` is not :class:`confluent_kafka.DeserializingConsumer`
        or :class:`kafkacrypto.KafkaConsumer`
    :raises ValueError: if `consumer_type` is :class:`kafkacrypto.KafkaConsumer` and
        `kafkacrypto` is None
    """

    def __init__(
        self,
        consumer_type,
        configs,
        kafkacrypto=None,
        message_key_regex=None,
        filter_new_message_keys=False,
        starting_offsets=None,
        filepath_regex=None,
        **kwargs,
    ):
        """
        Constructor method
        """
        super().__init__(**kwargs)
        if consumer_type == KafkaConsumer:
            if kafkacrypto is None:
                errmsg = "ERROR: creating a KafkaConsumer requires giving its KafkaCrypto objects!"
                self.logger.error(errmsg, exc_type=ValueError)
            self.__kafkacrypto = kafkacrypto
            self._consumer = consumer_type(**configs)
            self.__messages = []
        elif consumer_type == DeserializingConsumer:
            self._consumer = consumer_type(configs)
        else:
            errmsg = f"ERROR: Unrecognized consumer type {consumer_type} for OpenMSIStreamConsumer!"
            self.logger.error(errmsg, exc_type=ValueError)
        self.configs = configs
        self.message_key_regex = message_key_regex
        self.filter_new_message_keys = filter_new_message_keys
        self.filepath_regex = filepath_regex
        self.__starting_offsets = starting_offsets

    @staticmethod
    def get_consumer_args_kwargs(
        config_file_path, logger=None, kafkacrypto=None, **kwargs
    ):
        """
        Return the list of arguments and dictionary or keyword arguments that should be used
        to instantiate :class:`~OpenMSIStreamConsumer` objects based on the given config file.
        Used to share a single :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto` instance across
        several Consumers.

        :param config_file_path: Path to the config file to use in defining Consumers
        :type config_file_path: :class:`pathlib.Path`
        :param logger: The :class:`~.utilities.Logger` object to use for each of the
            :class:`~OpenMSIStreamConsumer` objects
        :type logger: :class:`~.utilities.Logger`
        :param kafkacrypto: The :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto` object
            that should be used to instantiate Consumers. Only needed if a single specific
            :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto` instance should be shared.
        :type kafkacrypto: :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto`, optional
        :param kwargs: Any extra keyword arguments are added to the configuration dict for
            the Consumers, with underscores in their names replaced by dots
        :type kwargs: dict

        :return: ret_args, the list of arguments to create new :class:`~OpenMSIStreamConsumer`
            objects based on the config file and arguments to this method
        :rtype: list
        :return: ret_kwargs, the dictionary of keyword arguments to create new
            :class:`~OpenMSIStreamConsumer` objects based on the config file and arguments
            to this method
        :rtype: dict
        """
        parser = KafkaConfigFileParser(config_file_path, logger=logger)
        ret_kwargs = {"logger": logger}
        # get the broker and consumer configurations
        all_configs = {**parser.broker_configs, **parser.consumer_configs}
        all_configs = add_kwargs_to_configs(all_configs, logger, **kwargs)
        # all_configs['debug']='broker,topic,msg'
        # if the group.id has been set as "create_new" generate a new group ID
        if "group.id" in all_configs and all_configs["group.id"].lower() == "create_new":
            all_configs["group.id"] = str(uuid.uuid1())
        # Use a DeserializingConsumer by default
        if kafkacrypto is None and parser.kc_config_file_str is None:
            ret_args = [DeserializingConsumer, all_configs]
            return ret_args, ret_kwargs
        # if "kafkacrypto" was given, or there are configs for KafkaCrypto, use a KafkaConsumer
        if kafkacrypto is not None:
            if not isinstance(kafkacrypto, OpenMSIStreamKafkaCrypto):
                errmsg = f"ERROR: kafkacrypto argument type = {type(kafkacrypto)}"
                raise_err_with_optional_logger(logger, errmsg, TypeError)
            debugmsg = (
                f"Consumed messages will be decrypted using configs "
                f"at {kafkacrypto.config_file_path}"
            )
            debug_msg_with_optional_logger(logger, debugmsg)
            k_c = kafkacrypto
        else:
            debugmsg = (
                f"Consumed messages will be decrypted using configs "
                f"at {parser.kc_config_file_str}"
            )
            debug_msg_with_optional_logger(logger, debugmsg)
            k_c = OpenMSIStreamKafkaCrypto(
                parser.broker_configs, parser.kc_config_file_str
            )
        if "key.deserializer" in all_configs:
            keydes = CompoundDeserializer(
                k_c.key_deserializer, all_configs.pop("key.deserializer")
            )
        else:
            keydes = k_c.key_deserializer
        if "value.deserializer" in all_configs:
            valdes = CompoundDeserializer(
                k_c.value_deserializer, all_configs.pop("value.deserializer")
            )
        else:
            valdes = k_c.value_deserializer
        all_configs["key_deserializer"] = keydes
        all_configs["value_deserializer"] = valdes
        ret_args = [KafkaConsumer, all_configs]
        ret_kwargs["kafkacrypto"] = k_c
        return ret_args, ret_kwargs

    @classmethod
    def from_file(cls, *args, **kwargs):
        """
        Wrapper around :func:`~OpenMSIStreamConsumer.get_consumer_args_kwargs` and the
        :class:`~OpenMSIStreamConsumer` constructor to return a single
        :class:`~OpenMSIStreamConsumer` from a given config file/arguments.
        Arguments are the same as :func:`~OpenMSIStreamConsumer.get_consumer_args_kwargs`

        :returns: An :class:`~OpenMSIStreamConsumer` object based on the given config file/arguments
        :rtype: :class:`~OpenMSIStreamConsumer`
        """
        args_to_use, kwargs_to_use = OpenMSIStreamConsumer.get_consumer_args_kwargs(
            *args, **kwargs
        )
        return cls(*args_to_use, **kwargs_to_use)

    def get_next_message(self, *poll_args, **poll_kwargs):
        """
        Call poll() for the underlying consumer and return any consumed message's value.

        If this Consumer's `message_key_regex` is not None this method may return None
        even though a message was consumed.

        For the case of encrypted messages, this may return a :class:`kafkacrypto.Message`
        object with :class:`kafkacrypto.KafkaCryptoMessages` for its key and/or value
        if the message fails to be decrypted within a certain amount of time

        All arguments/keyword arguments are passed through to the underlying Consumer's
        poll() function.

        :return: a single consumed :class:`confluent_kafka.Message` object, an undecrypted
            :class:`kafkacrypto.Message` object, or None
        :rtype: :class:`confluent_kafka.Message`, :class:`kafkacrypto.Message`, or None

        :raises Exception: for any error encountered in calling poll(). A special
            error-level message will be logged, including the stack trace of the Exception.
        """
        # There's one version for the result of a KafkaConsumer.poll() call
        if isinstance(self._consumer, KafkaConsumer):
            # check if there are any messages still waiting to be processed
            if len(self.__messages) > 0:
                consumed_msg = self.__messages.pop(0)
                return self._filter_message(consumed_msg)
            msg_dict = self._consumer.poll(*poll_args, **poll_kwargs)
            if msg_dict == {}:
                return None
            for t_p in msg_dict.keys():
                for msg in msg_dict[t_p]:
                    self.__messages.append(msg)
            return self.get_next_message(*poll_args, **poll_kwargs)
        # And another version for a regular Consumer
        consumed_msg = None
        try:
            consumed_msg = self._consumer.poll(*poll_args, **poll_kwargs)
        except Exception as exc:
            errmsg = (
                "ERROR: encountered an error in a call to consumer.poll() and this message "
                "will be skipped. Exception will be logged below and re-raised."
            )
            self.logger.error(errmsg, exc_info=exc, reraise=True)
            return None
        if consumed_msg is not None and consumed_msg != {}:
            if consumed_msg.error() is not None or consumed_msg.value() is None:
                warnmsg = (
                    f"WARNING: unexpected consumed message, consumed_msg = {consumed_msg}, "
                    f"consumed_msg.error() = {consumed_msg.error()}, "
                    f"consumed_msg.value() = {consumed_msg.value()}"
                )
                self.logger.warning(warnmsg)
            return self._filter_message(consumed_msg)
        return None

    def subscribe(self, *args, **kwargs):
        """
        A wrapper around the underlying Consumer's subscribe() method
        """
        return self._consumer.subscribe(*args, **kwargs)

    def commit(self, message=None, offsets=None, asynchronous=True):
        """
        A wrapper around the underlying Consumer's commit() method.
        Exactly one of `message` and `offsets` must be given.

        :param message: The message whose offset should be committed
        :type message: :class:`confluent_kafka.Message` or :class:`kafkacrypto.Message`, optional
        :param offsets: The list of topic+partitions+offsets to commit
        :type offsets: list(:class:`confluent_kafka.TopicPartition`), optional
        :param asynchronous: If True, the Consumer.commit call will return immediately and
            run asynchronously instead of blocking
        :type asynchronous: bool

        :raises ValueError: if `message` and `offsets` are both given
        """
        if (message is not None) and (not isinstance(message, Message)):
            try:
                offset_dict = {
                    KCCommitOffsetDictKey(
                        message.topic, message.partition
                    ): KCCommitOffset(message.offset)
                }
                if asynchronous:
                    return self._consumer.commit_async(offsets=offset_dict)
                return self._consumer.commit(offsets=offset_dict)
            except Exception:
                warnmsg = "WARNING: failed to commit an offset for an encrypted message. "
                warnmsg = "Duplicates may result if the Consumer is restarted."
                self.logger.warning(warnmsg)
                return None
        if message is None:
            return self._consumer.commit(offsets=offsets, asynchronous=asynchronous)
        if offsets is None:
            return self._consumer.commit(message=message, asynchronous=asynchronous)
        errmsg = (
            'ERROR: "message" and "offset" arguments are exclusive for Consumer.commit. '
            "Nothing committed."
        )
        self.logger.error(errmsg, exc_type=ValueError)
        return None

    def close(self, *args, **kwargs):
        """
        Combined wrapper around the underlying Consumer's close() method and
        :func:`kafkacrypto.KafkaCrypto.close`.
        """
        self._consumer.close(*args, **kwargs)
        try:
            if self.__kafkacrypto:
                self.__kafkacrypto.close()
        except Exception:
            pass
        finally:
            self.__kafkacrypto = None

    def _filter_message(self, msg):
        """
        Checks a message's key against the key regex, its filepath against the filepath regex,
        and its offset against the starting offsets. Returns None if a message should be skipped,
        otherwise returns the message.
        """
        if self._message_passes_key_filter(msg) and self._message_passes_filepath_filter(
            msg
        ):
            return msg
        return None

    def _message_passes_key_filter(self, msg):
        """
        Returns True if a message should be consumed based on the key regex and False
        if it should be skipped instead
        """
        if (self.message_key_regex is None) or not (
            self.filter_new_message_keys or self.message_consumed_before(msg)
        ):
            return True
        try:
            msg_key = msg.key()  # from a regular Kafka Consumer
        except TypeError:
            msg_key = msg.key  # from KafkaCrypto
        if not isinstance(msg_key, str):
            warnmsg = (
                f"WARNING: found a message whose key ({msg_key}) is not a string, "
                "but which should be filtered using the key regex. "
                "This message will be consumed as though it successfully passed the filter."
            )
            self.logger.warning(warnmsg)
            return True
        if self.message_key_regex.match(msg_key):
            return True
        return False

    def _message_passes_filepath_filter(self, msg):
        """
        Returns True if a message should be consumed based on the filepath regex and False
        if it should be skipped instead
        """
        if self.filepath_regex is None:
            return True
        try:
            msg_value = msg.value()  # from a regular Kafka Consumer
        except TypeError:
            msg_value = msg.value  # from KafkaCrypto
        if not isinstance(msg_value, DataFileChunk):
            warnmsg = (
                f"WARNING: found a message whose value ({msg_value}) is not a DataFileChunk, "
                "but which should be filtered using the filepath regex. "
                "This message will be consumed as though it successfully passed the filter."
            )
            self.logger.warning(warnmsg)
            return True
        rel_filepath = msg_value.relative_filepath
        if self.filepath_regex.match(str(rel_filepath)):
            return True
        return False

    @methodtools.lru_cache()
    def message_consumed_before(self, msg):
        """
        Returns True if a message has an offset less than the starting offset
        for the topic/partition, False otherwise
        """
        try:  # from a regular Kafka Consumer
            msg_topic = msg.topic()
            msg_partition = msg.partition()
            msg_offset = msg.offset()
        except TypeError:  # from KafkaCrypto
            msg_topic = msg.topic
            msg_partition = msg.partition
            msg_offset = msg.offset
        if not (
            isinstance(msg_topic, str)
            and isinstance(msg_partition, int)
            and isinstance(msg_offset, int)
        ):
            errmsg = (
                "ERROR: failed to check whether a message has been consumed before "
                f"because its topic/partition/offset are {msg_topic}/{msg_partition}/{msg_offset}!"
            )
            self.logger.error(errmsg, exc_type=TypeError)
        starting_offset = None
        for tpo in self.__starting_offsets:
            if tpo.topic == msg_topic and tpo.partition == msg_partition:
                starting_offset = tpo.offset
        if starting_offset is None:
            errmsg = (
                "ERROR: failed to check whether a message has been consumed before "
                "because its topic/partition were not found in the list of starting offsets"
            )
            self.logger.error(errmsg, exc_type=ValueError)
        return msg_offset < starting_offset



"""Helper classes/wrappers/infrastructure used throughout OpenMSIStream"""

from .argument_parsing import OpenMSIStreamArgumentParser
from .controlled_process import ControlledProcess
from .controlled_process_single_thread import ControlledProcessSingleThread
from .controlled_process_multi_threaded import ControlledProcessMultiThreaded
from .dataclass_table import (
    DataclassTableReadOnly,
    DataclassTableAppendOnly,
    DataclassTable,
)
from .has_arguments import HasArguments
from .has_argument_parser import HasArgumentParser
from .logging import Logger, LogOwner
from .runnable import Runnable

__all__ = [
    "OpenMSIStreamArgumentParser",
    "ControlledProcess",
    "ControlledProcessSingleThread",
    "ControlledProcessMultiThreaded",
    "DataclassTableReadOnly",
    "DataclassTableAppendOnly",
    "DataclassTable",
    "HasArguments",
    "HasArgumentParser",
    "Logger",
    "LogOwner",
    "Runnable",
]



"""Miscellaneous functions/classes used elsewhere in the Kafka wrapper"""

# imports
import warnings
from collections import namedtuple
from confluent_kafka import OFFSET_BEGINNING

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto import KafkaConsumer


def add_kwargs_to_configs(configs, logger, **kwargs):
    """
    Add any kwargs with underscores replaced with dots to a given config dictionary
    """
    new_configs = configs.copy()
    for argname, arg in kwargs.items():
        config_name = argname.replace("_", ".")
        if config_name in new_configs and new_configs[config_name] != arg:
            if logger is not None:
                msg = (
                    f'A new value for the "{config_name}" config has been supplied by '
                    f"a keyword argument that will overwrite a previously-set value. "
                    f"The value will be set to {arg} instead of {new_configs[config_name]}."
                )
                logger.debug(msg)
        new_configs[config_name] = arg
    return new_configs


def default_producer_callback(err, msg, prodid, logger=None, **other_kwargs):
    """
    Default callback function to use for testing whether a message has been
    successfully produced to the topic
    """
    if err is None and msg.error() is not None:
        err = msg.error()
    if err is not None:  # raise an error if the message wasn't sent successfully
        if err.fatal():
            logmsg = (
                f"ERROR: Producer with ID {prodid} fatally failed to deliver message "
                f'with kwargs "{other_kwargs}". MESSAGE DROPPED. Error reason: {err.str()}'
            )
            if logger is not None:
                logger.error(logmsg)
            else:
                raise RuntimeError(logmsg)
        elif not err.retriable():
            logmsg = (
                f"ERROR: Producer with ID {prodid} failed to deliver message with kwargs "
                f'"{other_kwargs}" and cannot retry. MESSAGE DROPPED. '
                f"Error reason: {err.str()}"
            )
            if logger is not None:
                logger.error(logmsg)
            else:
                raise RuntimeError(logmsg)


def make_callback(func, *args, **kwargs):
    """
    Callbacks are rendered at runtime and so regular functions used as producer callbacks
    need to be wrapped in this lambda
    """
    return lambda err, msg: func(err, msg, *args, **kwargs)


KCCommitOffsetDictKey = namedtuple("KCCommitOffsetDictKey", ["topic", "partition"])
KCCommitOffset = namedtuple("KCCommitOffset", ["offset"])


def reset_to_beginning_on_assign(consumer, partitions):
    """
    Used as the "on_assign" parameter to a Consumer constructor to get each partition
    to start at its earliest message when it's assigned to a Consumer
    """
    for part in partitions:
        part.offset = OFFSET_BEGINNING
    if isinstance(consumer, KafkaConsumer):
        consumer.assign_and_seek(partitions)
    else:
        consumer.assign(partitions)
